import contextlib
import dataclasses
import subprocess
import sys
import textwrap
from collections import OrderedDict
from collections.abc import Iterable, Iterator, Sequence, Set
from pathlib import Path, PurePath, PurePosixPath
from typing import TYPE_CHECKING, assert_never

from cibuildwheel import errors
from cibuildwheel.architecture import Architecture
from cibuildwheel.frontend import get_build_frontend_extra_flags
from cibuildwheel.logger import log
from cibuildwheel.oci_container import OCIContainer, OCIContainerEngineConfig, OCIPlatform
from cibuildwheel.options import BuildOptions, Options
from cibuildwheel.selector import BuildSelector
from cibuildwheel.util import resources
from cibuildwheel.util.file import copy_test_sources
from cibuildwheel.util.helpers import prepare_command, unwrap
from cibuildwheel.util.packaging import find_compatible_wheel

if TYPE_CHECKING:
    from cibuildwheel.typing import PathOrStr

ARCHITECTURE_OCI_PLATFORM_MAP = {
    Architecture.x86_64: OCIPlatform.AMD64,
    Architecture.i686: OCIPlatform.i386,
    Architecture.aarch64: OCIPlatform.ARM64,
    Architecture.ppc64le: OCIPlatform.PPC64LE,
    Architecture.s390x: OCIPlatform.S390X,
    Architecture.armv7l: OCIPlatform.ARMV7,
    Architecture.riscv64: OCIPlatform.RISCV64,
}


@dataclasses.dataclass(frozen=True, kw_only=True)
class PythonConfiguration:
    version: str
    identifier: str
    path_str: str

    @property
    def path(self) -> PurePosixPath:
        return PurePosixPath(self.path_str)


@dataclasses.dataclass(frozen=True, kw_only=True)
class BuildStep:
    platform_configs: list[PythonConfiguration]
    platform_tag: str
    container_engine: OCIContainerEngineConfig
    container_image: str


def all_python_configurations() -> list[PythonConfiguration]:
    config_dicts = resources.read_python_configs("linux")
    return [PythonConfiguration(**item) for item in config_dicts]


def get_python_configurations(
    build_selector: BuildSelector,
    architectures: Set[Architecture],
) -> list[PythonConfiguration]:
    python_configurations = all_python_configurations()

    # return all configurations whose arch is in our `architectures` set,
    # and match the build/skip rules
    return [
        c
        for c in python_configurations
        if any(c.identifier.endswith(arch.value) for arch in architectures)
        and build_selector(c.identifier)
    ]


def container_image_for_python_configuration(
    config: PythonConfiguration, build_options: BuildOptions
) -> str:
    # e.g
    # identifier is 'cp310-manylinux_x86_64'
    # platform_tag is 'manylinux_x86_64'
    # platform_arch is 'x86_64'
    _, platform_tag = config.identifier.split("-", 1)
    _, platform_arch = platform_tag.split("_", 1)

    assert build_options.manylinux_images is not None
    assert build_options.musllinux_images is not None

    return (
        build_options.manylinux_images[platform_arch]
        if platform_tag.startswith("manylinux")
        else build_options.musllinux_images[platform_arch]
    )


def get_build_steps(
    options: Options, python_configurations: list[PythonConfiguration]
) -> Iterator[BuildStep]:
    """
    Groups PythonConfigurations into BuildSteps. Each BuildStep represents a
    separate container instance.
    """
    steps = OrderedDict[tuple[str, str, str, OCIContainerEngineConfig], BuildStep]()

    for config in python_configurations:
        _, platform_tag = config.identifier.split("-", 1)

        build_options = options.build_options(config.identifier)

        before_all = build_options.before_all
        container_image = container_image_for_python_configuration(config, build_options)
        container_engine = build_options.container_engine

        step_key = (platform_tag, container_image, before_all, container_engine)

        if step_key in steps:
            steps[step_key].platform_configs.append(config)
        else:
            steps[step_key] = BuildStep(
                platform_configs=[config],
                platform_tag=platform_tag,
                container_engine=container_engine,
                container_image=container_image,
            )

    yield from steps.values()


def check_all_python_exist(
    *, platform_configs: Iterable[PythonConfiguration], container: OCIContainer
) -> None:
    exist = True
    has_manylinux_interpreters = False
    messages = []

    with contextlib.suppress(subprocess.CalledProcessError):
        # use capture_output to keep quiet
        container.call(["manylinux-interpreters", "--help"], capture_output=True)
        has_manylinux_interpreters = True

    for config in platform_configs:
        python_path = config.path / "bin" / "python"
        if has_manylinux_interpreters:
            try:
                container.call(["manylinux-interpreters", "ensure", config.path.name])
            except subprocess.CalledProcessError:
                messages.append(
                    f"  'manylinux-interpreters ensure {config.path.name}' needed to build '{config.identifier}' failed in container running image '{container.image}'."
                    " Either the installation failed or this interpreter is not available in that image. Please check the logs."
                )
                exist = False
        else:
            try:
                container.call(["test", "-x", python_path])
            except subprocess.CalledProcessError:
                messages.append(
                    f"  '{python_path}' executable doesn't exist in image '{container.image}' to build '{config.identifier}'."
                )
                exist = False
    if not exist:
        message = "\n".join(messages)
        raise errors.FatalError(message)


def build_in_container(
    *,
    options: Options,
    platform_configs: Sequence[PythonConfiguration],
    container: OCIContainer,
    container_project_path: PurePath,
    container_package_dir: PurePath,
    local_tmp_dir: Path,
) -> None:
    container_output_dir = PurePosixPath("/output")

    check_all_python_exist(platform_configs=platform_configs, container=container)

    log.step("Copying project into container...")
    container.copy_into(Path.cwd(), container_project_path)

    before_all_options_identifier = platform_configs[0].identifier
    before_all_options = options.build_options(before_all_options_identifier)

    if before_all_options.before_all:
        log.step("Running before_all...")

        env = container.get_environment()
        env["PATH"] = f"/opt/python/cp39-cp39/bin:{env['PATH']}"
        env["PIP_DISABLE_PIP_VERSION_CHECK"] = "1"
        env["PIP_ROOT_USER_ACTION"] = "ignore"
        env = before_all_options.environment.as_dictionary(
            env, executor=container.environment_executor
        )

        before_all_prepared = prepare_command(
            before_all_options.before_all,
            project=container_project_path,
            package=container_package_dir,
        )
        container.call(["sh", "-c", before_all_prepared], env=env)

    built_wheels: list[PurePosixPath] = []

    for config in platform_configs:
        log.build_start(config.identifier)
        local_identifier_tmp_dir = local_tmp_dir / config.identifier
        build_options = options.build_options(config.identifier)
        build_frontend = build_options.build_frontend
        use_uv = build_frontend.name in {"build[uv]", "uv"}
        pip = ["uv", "pip"] if use_uv else ["pip"]

        log.step("Setting up build environment...")

        dependency_constraint_flags: list[PathOrStr] = []
        local_constraints_file = build_options.dependency_constraints.get_for_python_version(
            version=config.version,
            tmp_dir=local_identifier_tmp_dir,
        )
        if local_constraints_file:
            container_constraints_file = PurePosixPath("/constraints.txt")
            container.copy_into(local_constraints_file, container_constraints_file)
            dependency_constraint_flags = ["-c", container_constraints_file]

        env = container.get_environment()
        env["PIP_DISABLE_PIP_VERSION_CHECK"] = "1"
        env["PIP_ROOT_USER_ACTION"] = "ignore"

        # put this config's python top of the list
        python_bin = config.path / "bin"
        env["PATH"] = f"{python_bin}:{env['PATH']}"

        env = build_options.environment.as_dictionary(env, executor=container.environment_executor)

        # check config python is still on PATH
        which_python = container.call(["which", "python"], env=env, capture_output=True).strip()
        if PurePosixPath(which_python) != python_bin / "python":
            msg = "python available on PATH doesn't match our installed instance. If you have modified PATH, ensure that you don't overwrite cibuildwheel's entry or insert python above it."
            raise errors.FatalError(msg)

        if use_uv:
            which_uv = container.call(["which", "uv"], env=env, capture_output=True).strip()
            if not which_uv:
                msg = "uv not found on PATH. You must use a supported manylinux or musllinux environment with uv."
                raise errors.FatalError(msg)
        else:
            which_pip = container.call(["which", "pip"], env=env, capture_output=True).strip()
            if PurePosixPath(which_pip) != python_bin / "pip":
                msg = "pip available on PATH doesn't match our installed instance. If you have modified PATH, ensure that you don't overwrite cibuildwheel's entry or insert pip above it."
                raise errors.FatalError(msg)

        compatible_wheel = find_compatible_wheel(built_wheels, config.identifier)
        if compatible_wheel:
            log.step_end()
            print(
                f"\nFound previously built wheel {compatible_wheel.name}, that's compatible with {config.identifier}. Skipping build step..."
            )
            repaired_wheel = compatible_wheel
        else:
            if build_options.before_build:
                log.step("Running before_build...")
                before_build_prepared = prepare_command(
                    build_options.before_build,
                    project=container_project_path,
                    package=container_package_dir,
                )
                container.call(["sh", "-c", before_build_prepared], env=env)

            log.step("Building wheel...")

            temp_dir = PurePosixPath("/tmp/cibuildwheel")
            built_wheel_dir = temp_dir / "built_wheel"
            container.call(["rm", "-rf", built_wheel_dir])
            container.call(["mkdir", "-p", built_wheel_dir])

            extra_flags = get_build_frontend_extra_flags(
                build_frontend,
                build_options.build_verbosity,
                build_options.config_settings,
                py38=config.identifier[1:].startswith("p38"),
            )

            match build_frontend.name:
                case "pip":
                    container.call(
                        [
                            "python",
                            "-m",
                            "pip",
                            "wheel",
                            container_package_dir,
                            f"--wheel-dir={built_wheel_dir}",
                            "--no-deps",
                            *extra_flags,
                        ],
                        env=env,
                    )
                case "build" | "build[uv]":
                    if use_uv and "--no-isolation" not in extra_flags and "-n" not in extra_flags:
                        extra_flags += ["--installer=uv"]
                    container.call(
                        [
                            "python",
                            "-m",
                            "build",
                            container_package_dir,
                            "--wheel",
                            f"--outdir={built_wheel_dir}",
                            *extra_flags,
                        ],
                        env=env,
                    )
                case "uv":
                    container.call(
                        [
                            "uv",
                            "build",
                            f"--python={python_bin / 'python'}",
                            container_package_dir,
                            "--wheel",
                            f"--out-dir={built_wheel_dir}",
                            *extra_flags,
                        ],
                        env=env,
                    )
                case _:
                    assert_never(build_frontend)

            built_wheel = container.glob(built_wheel_dir, "*.whl")[0]

            repaired_wheel_dir = temp_dir / "repaired_wheel"
            container.call(["rm", "-rf", repaired_wheel_dir])
            container.call(["mkdir", "-p", repaired_wheel_dir])

            if built_wheel.name.endswith("none-any.whl"):
                raise errors.NonPlatformWheelError()

            if build_options.repair_command:
                log.step("Repairing wheel...")
                repair_command_prepared = prepare_command(
                    build_options.repair_command,
                    wheel=built_wheel,
                    dest_dir=repaired_wheel_dir,
                    package=container_package_dir,
                    project=container_project_path,
                )
                container.call(["sh", "-c", repair_command_prepared], env=env)
            else:
                container.call(["mv", built_wheel, repaired_wheel_dir])

            match container.glob(repaired_wheel_dir, "*.whl"):
                case []:
                    raise errors.RepairStepProducedNoWheelError()
                case [repaired_wheel]:
                    pass
                case too_many:
                    raise errors.RepairStepProducedMultipleWheelsError([p.name for p in too_many])

            if repaired_wheel.name in {wheel.name for wheel in built_wheels}:
                raise errors.AlreadyBuiltWheelError(repaired_wheel.name)

        if build_options.test_command and build_options.test_selector(config.identifier):
            log.step("Testing wheel...")

            # set up a virtual environment to install and test from, to make sure
            # there are no dependencies that were pulled in at build time.
            if not use_uv:
                container.call(
                    ["pip", "install", "virtualenv", *dependency_constraint_flags], env=env
                )

            testing_temp_dir = PurePosixPath(
                container.call(["mktemp", "-d"], capture_output=True).strip()
            )
            venv_dir = testing_temp_dir / "venv"

            if use_uv:
                container.call(["uv", "venv", venv_dir, "--python", python_bin / "python"], env=env)
            else:
                # Use embedded dependencies from virtualenv to ensure determinism
                venv_args = ["--no-periodic-update", "--pip=embed", "--no-setuptools"]
                if "38" in config.identifier:
                    venv_args.append("--no-wheel")
                container.call(["python", "-m", "virtualenv", *venv_args, venv_dir], env=env)

            virtualenv_env = env.copy()
            virtualenv_env["PATH"] = f"{venv_dir / 'bin'}:{virtualenv_env['PATH']}"
            virtualenv_env["VIRTUAL_ENV"] = str(venv_dir)
            virtualenv_env = build_options.test_environment.as_dictionary(
                prev_environment=virtualenv_env
            )

            if build_options.before_test:
                before_test_prepared = prepare_command(
                    build_options.before_test,
                    project=container_project_path,
                    package=container_package_dir,
                )
                container.call(["sh", "-c", before_test_prepared], env=virtualenv_env)

            # Install the wheel we just built
            container.call(
                [*pip, "install", str(repaired_wheel) + build_options.test_extras],
                env=virtualenv_env,
            )

            # Install any requirements to run the tests
            if build_options.test_requires:
                container.call([*pip, "install", *build_options.test_requires], env=virtualenv_env)

            # Run the tests from a different directory
            test_command_prepared = prepare_command(
                build_options.test_command,
                project=container_project_path,
                package=container_package_dir,
                wheel=repaired_wheel,
            )

            test_cwd = testing_temp_dir / "test_cwd"
            container.call(["mkdir", "-p", test_cwd])

            if build_options.test_sources:
                copy_test_sources(
                    build_options.test_sources,
                    Path.cwd(),
                    test_cwd,
                    copy_into=container.copy_into,
                )
            else:
                # Use the test_fail.py file to raise a nice error if the user
                # tries to run tests in the cwd
                container.copy_into(resources.TEST_FAIL_CWD_FILE, test_cwd / "test_fail.py")

            container.call(["sh", "-c", test_command_prepared], cwd=test_cwd, env=virtualenv_env)

            # clean up test environment
            container.call(["rm", "-rf", testing_temp_dir])

        # move repaired wheel to output
        output_wheel: Path | None = None
        if compatible_wheel is None:
            container.call(["mkdir", "-p", container_output_dir])
            container.call(["mv", repaired_wheel, container_output_dir])
            built_wheels.append(container_output_dir / repaired_wheel.name)
            output_wheel = options.globals.output_dir / repaired_wheel.name

        log.build_end(output_wheel)

    log.step("Copying wheels back to host...")
    # copy the output back into the host
    container.copy_out(container_output_dir, options.globals.output_dir)
    log.step_end()


def build(options: Options, tmp_path: Path) -> None:
    python_configurations = get_python_configurations(
        options.globals.build_selector, options.globals.architectures
    )

    cwd = Path.cwd()
    abs_package_dir = options.globals.package_dir.resolve()
    if cwd != abs_package_dir and cwd not in abs_package_dir.parents:
        msg = "package_dir must be inside the working directory"
        raise Exception(msg)

    container_project_path = PurePosixPath("/project")
    container_package_dir = container_project_path / abs_package_dir.relative_to(cwd)

    for build_step in get_build_steps(options, python_configurations):
        try:
            # check the container engine is installed
            subprocess.run(
                [build_step.container_engine.name, "--version"],
                check=True,
                stdout=subprocess.DEVNULL,
            )
        except subprocess.CalledProcessError as error:
            msg = unwrap(
                f"""
                {build_step.container_engine.name} not found. An OCI exe like
                Docker or Podman is required to run Linux builds. If you're
                building on Travis CI, add `services: [docker]` to your
                .travis.yml. If you're building on Circle CI in Linux, add a
                `setup_remote_docker` step to your .circleci/config.yml. If
                you're building on Cirrus CI, use `docker_builder` task.
                """
            )
            raise errors.ConfigurationError(msg) from error

        try:
            ids_to_build = [x.identifier for x in build_step.platform_configs]
            log.step(f"Starting container image {build_step.container_image}...")

            print(f"info: This container will host the build for {', '.join(ids_to_build)}...")
            architecture = Architecture(build_step.platform_tag.split("_", 1)[1])

            with OCIContainer(
                image=build_step.container_image,
                oci_platform=ARCHITECTURE_OCI_PLATFORM_MAP[architecture],
                cwd=container_project_path,
                engine=build_step.container_engine,
            ) as container:
                build_in_container(
                    options=options,
                    platform_configs=build_step.platform_configs,
                    container=container,
                    container_project_path=container_project_path,
                    container_package_dir=container_package_dir,
                    local_tmp_dir=tmp_path,
                )

        except subprocess.CalledProcessError as error:
            troubleshoot(options, error)
            msg = f"Command {error.cmd} failed with code {error.returncode}. {error.stdout or ''}"
            raise errors.FatalError(msg) from error


def _matches_prepared_command(error_cmd: Sequence[str], command_template: str) -> bool:
    if len(error_cmd) < 3 or error_cmd[0:2] != ["sh", "-c"]:
        return False
    command_prefix = command_template.split("{", maxsplit=1)[0].strip()
    return error_cmd[2].startswith(command_prefix)


def troubleshoot(options: Options, error: Exception) -> None:
    if isinstance(error, subprocess.CalledProcessError) and (
        error.cmd[0:4] == ["python", "-m", "pip", "wheel"]
        or error.cmd[0:2] == ["uv", "build"]
        or error.cmd[0:3] == ["python", "-m", "build"]
        or _matches_prepared_command(
            error.cmd, options.build_options(None).repair_command
        )  # TODO allow matching of overrides too?
    ):
        # the wheel build step or the repair step failed
        so_files = list(options.globals.package_dir.glob("**/*.so"))

        if so_files:
            print(
                textwrap.dedent(
                    """
                    NOTE: Shared object (.so) files found in this project.

                    These files might be built against the wrong OS, causing problems with
                    auditwheel. If possible, run cibuildwheel in a clean checkout.

                    If you're using Cython and have previously done an in-place build,
                    remove those build files (*.so and *.c) before starting cibuildwheel.

                    setuptools uses the build/ folder to store its build cache. It
                    may be necessary to remove those build files (*.so and *.o) before
                    starting cibuildwheel.

                    Files that belong to a virtual environment are probably not an issue
                    unless you used a custom command telling cibuildwheel to activate it.
                    """
                ),
                file=sys.stderr,
            )

            print("  Files detected:")
            print("\n".join(f"    {f}" for f in so_files))
            print()



import ssl
from pathlib import Path

import certifi
import pytest

from cibuildwheel.util.file import download

DOWNLOAD_URL = "https://cdn.jsdelivr.net/gh/pypa/cibuildwheel@v1.6.3/requirements-dev.txt"


def test_download(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.delenv("SSL_CERT_FILE", raising=False)
    dest = tmp_path / "file.txt"
    download(DOWNLOAD_URL, dest)
    assert len(dest.read_bytes()) == 134


def test_download_good_ssl_cert_file(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setenv("SSL_CERT_FILE", certifi.where())
    dest = tmp_path / "file.txt"
    download(DOWNLOAD_URL, dest)
    assert len(dest.read_bytes()) == 134


def test_download_bad_ssl_cert_file(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    bad_cafile = tmp_path / "ca.pem"
    bad_cafile.write_text("bad certificates")
    monkeypatch.setenv("SSL_CERT_FILE", str(bad_cafile))
    dest = tmp_path / "file.txt"
    with pytest.raises(ssl.SSLError):
        download(DOWNLOAD_URL, dest)



import contextlib
import platform as platform_module
import subprocess
import sys
from collections.abc import Generator, Iterator
from pathlib import Path
from typing import Any

import pytest

from cibuildwheel import architecture
from cibuildwheel.logger import Logger
from cibuildwheel.platforms import android, ios, linux, macos, pyodide, windows
from cibuildwheel.util import file


class ArgsInterceptor:
    def __init__(self) -> None:
        self.call_count = 0
        self.args: tuple[Any, ...] = ()
        self.kwargs: dict[str, Any] = {}

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        self.call_count += 1
        self.args = args
        self.kwargs = kwargs


@pytest.fixture(autouse=True)
def mock_protection(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Ensure that a unit test will never actually run a cibuildwheel 'build'
    function, which shouldn't be run on a developer's machine
    """

    def fail_on_call(*args: object, **kwargs: object) -> None:
        msg = "This should never be called"
        raise RuntimeError(msg)

    def ignore_call(*args: object, **kwargs: object) -> None:
        pass

    monkeypatch.setattr(subprocess, "Popen", fail_on_call)
    monkeypatch.setattr(file, "download", fail_on_call)
    monkeypatch.setattr(windows, "build", fail_on_call)
    monkeypatch.setattr(linux, "build", fail_on_call)
    monkeypatch.setattr(macos, "build", fail_on_call)
    monkeypatch.setattr(pyodide, "build", fail_on_call)
    monkeypatch.setattr(Path, "mkdir", ignore_call)
    monkeypatch.setattr(architecture, "_check_aarch32_el0", lambda: True)


@pytest.fixture(autouse=True)
def fake_package_dir_autouse(fake_package_dir: list[str]) -> None:
    pass


@pytest.fixture(autouse=True)
def disable_print_wheels(monkeypatch: pytest.MonkeyPatch) -> None:
    @contextlib.contextmanager
    def empty_cm(*args: object, **kwargs: object) -> Generator[None, None, None]:
        yield

    monkeypatch.setattr(Logger, "print_summary", empty_cm)


@pytest.fixture
def allow_empty(monkeypatch: pytest.MonkeyPatch, fake_package_dir: list[str]) -> None:
    monkeypatch.setattr(sys, "argv", [*fake_package_dir, "--allow-empty"])


@pytest.fixture(params=["linux", "macos", "windows"])
def platform(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch) -> str:
    platform_value: str = request.param
    monkeypatch.setenv("CIBW_PLATFORM", platform_value)

    if platform_value == "windows":
        monkeypatch.setattr(platform_module, "machine", lambda: "AMD64")
    else:
        monkeypatch.setattr(platform_module, "machine", lambda: "x86_64")

    if platform_value == "macos":
        monkeypatch.setattr(macos, "get_macos_version", lambda: (11, 1))

    return platform_value


@pytest.fixture
def intercepted_build_args(monkeypatch: pytest.MonkeyPatch) -> Iterator[ArgsInterceptor]:
    intercepted = ArgsInterceptor()

    monkeypatch.setattr(android, "build", intercepted)
    monkeypatch.setattr(ios, "build", intercepted)
    monkeypatch.setattr(linux, "build", intercepted)
    monkeypatch.setattr(macos, "build", intercepted)
    monkeypatch.setattr(pyodide, "build", intercepted)
    monkeypatch.setattr(windows, "build", intercepted)

    yield intercepted

    # check that intercepted_build_args only ever had one set of args
    assert intercepted.call_count <= 1



import os
import sys

from cibuildwheel.environment import parse_environment

# this command is equivalent to Unix 'echo', but works on Windows too
PYTHON_ECHO = f"'{sys.executable}' -c \"import sys; print(*sys.argv[1:])\""


def test_basic_parsing() -> None:
    environment_recipe = parse_environment("VAR=1 VBR=2")

    environment_dict = environment_recipe.as_dictionary(prev_environment={})

    assert environment_dict == {"VAR": "1", "VBR": "2"}


def test_quotes() -> None:
    environment_recipe = parse_environment("A=1 VAR=\"1 NOT_A_VAR=2\" VBR='vbr'")

    environment_dict = environment_recipe.as_dictionary(prev_environment={})

    assert environment_dict == {"A": "1", "VAR": "1 NOT_A_VAR=2", "VBR": "vbr"}


def test_inheritance() -> None:
    environment_recipe = parse_environment("PATH=$PATH:/usr/local/bin")

    environment_dict = environment_recipe.as_dictionary(prev_environment={"PATH": "/usr/bin"})

    assert environment_dict == {"PATH": "/usr/bin:/usr/local/bin"}


def test_shell_eval() -> None:
    environment_recipe = parse_environment(f'VAR="$({PYTHON_ECHO} "a   test" string)"')

    env_copy = os.environ.copy()
    env_copy.pop("VAR", None)

    environment_dict = environment_recipe.as_dictionary(prev_environment=env_copy)

    assert environment_dict["VAR"] == "a   test string"


def test_shell_eval_and_env() -> None:
    environment_recipe = parse_environment(f'VAR="$({PYTHON_ECHO} "$PREV_VAR" string)"')

    prev_environment = {**os.environ, "PREV_VAR": "1 2 3"}
    environment_dict = environment_recipe.as_dictionary(prev_environment=prev_environment)

    assert environment_dict == {**prev_environment, "VAR": "1 2 3 string"}


def test_empty_var() -> None:
    environment_recipe = parse_environment("CFLAGS=")

    environment_dict = environment_recipe.as_dictionary(prev_environment={"CFLAGS": "-Wall"})

    assert environment_dict == {"CFLAGS": ""}


def test_no_vars() -> None:
    environment_recipe = parse_environment("")

    environment_dict = environment_recipe.as_dictionary(prev_environment={})

    assert environment_dict == {}


def test_no_vars_pass_through() -> None:
    environment_recipe = parse_environment("")

    environment_dict = environment_recipe.as_dictionary(
        prev_environment={"CIBUILDWHEEL": "awesome"}
    )

    assert environment_dict == {"CIBUILDWHEEL": "awesome"}


def test_operators_inside_eval() -> None:
    environment_recipe = parse_environment(
        f'SOMETHING="$({PYTHON_ECHO} a; {PYTHON_ECHO} b; {PYTHON_ECHO} c)"'
    )

    # pass the existing process env so subcommands can be run in the evaluation
    environment_dict = environment_recipe.as_dictionary(prev_environment=os.environ.copy())

    assert environment_dict.get("SOMETHING") == "a\nb\nc"


def test_substitution_with_backslash() -> None:
    environment_recipe = parse_environment('PATH2="somewhere_else;$PATH1"')

    environment_dict = environment_recipe.as_dictionary(prev_environment={"PATH1": "c:\\folder\\"})

    assert environment_dict.get("PATH2") == "somewhere_else;c:\\folder\\"


def test_awkwardly_quoted_variable() -> None:
    environment_recipe = parse_environment(
        f'VAR2=something"like this""$VAR1"$VAR1$({PYTHON_ECHO} "there is more")"$({PYTHON_ECHO} "and more!")"'
    )

    prev_environment = {**os.environ, "VAR1": "but wait"}
    environment_dict = environment_recipe.as_dictionary(prev_environment=prev_environment)

    assert (
        environment_dict.get("VAR2") == "somethinglike thisbut waitbut waitthere is moreand more!"
    )



import textwrap
from pathlib import Path
from pprint import pprint

import pytest

import cibuildwheel.platforms.linux
from cibuildwheel.oci_container import OCIContainerEngineConfig
from cibuildwheel.options import CommandLineArguments, Options


def test_linux_container_split(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Tests splitting linux builds by container image, container engine, and before_all
    """

    args = CommandLineArguments.defaults()
    args.platform = "linux"

    (tmp_path / "pyproject.toml").write_text(
        textwrap.dedent(
            """
                [tool.cibuildwheel]
                manylinux-x86_64-image = "normal_container_image"
                manylinux-i686-image = "normal_container_image"
                build = "*-manylinux_x86_64"
                skip = "[gp]p*"
                archs = "x86_64 i686"

                [[tool.cibuildwheel.overrides]]
                select = "cp{38,39,310}-*"
                manylinux-x86_64-image = "other_container_image"
                manylinux-i686-image = "other_container_image"

                [[tool.cibuildwheel.overrides]]
                select = "cp39-*"
                before-all = "echo 'a cp39-only command'"

                [[tool.cibuildwheel.overrides]]
                select = "cp310-*"
                container-engine = "docker; create_args: --privileged"
            """
        )
    )

    monkeypatch.chdir(tmp_path)
    options = Options("linux", command_line_arguments=args, env={})

    python_configurations = cibuildwheel.platforms.linux.get_python_configurations(
        options.globals.build_selector, options.globals.architectures
    )

    build_steps = list(cibuildwheel.platforms.linux.get_build_steps(options, python_configurations))

    # helper functions to extract test info
    def identifiers(step: cibuildwheel.platforms.linux.BuildStep) -> list[str]:
        return [c.identifier for c in step.platform_configs]

    def before_alls(step: cibuildwheel.platforms.linux.BuildStep) -> list[str]:
        return [options.build_options(c.identifier).before_all for c in step.platform_configs]

    def container_engines(
        step: cibuildwheel.platforms.linux.BuildStep,
    ) -> list[OCIContainerEngineConfig]:
        return [options.build_options(c.identifier).container_engine for c in step.platform_configs]

    pprint(build_steps)

    default_container_engine = OCIContainerEngineConfig(name="docker")

    assert build_steps[0].container_image == "other_container_image"
    assert identifiers(build_steps[0]) == ["cp38-manylinux_x86_64"]
    assert before_alls(build_steps[0]) == [""]
    assert container_engines(build_steps[0]) == [default_container_engine]

    assert build_steps[1].container_image == "other_container_image"
    assert identifiers(build_steps[1]) == ["cp39-manylinux_x86_64"]
    assert before_alls(build_steps[1]) == ["echo 'a cp39-only command'"]
    assert container_engines(build_steps[1]) == [default_container_engine]

    assert build_steps[2].container_image == "other_container_image"
    assert identifiers(build_steps[2]) == ["cp310-manylinux_x86_64"]
    assert before_alls(build_steps[2]) == [""]
    assert container_engines(build_steps[2]) == [
        OCIContainerEngineConfig(name="docker", create_args=("--privileged",))
    ]

    assert build_steps[3].container_image == "normal_container_image"
    assert identifiers(build_steps[3]) == [
        "cp311-manylinux_x86_64",
        "cp312-manylinux_x86_64",
        "cp313-manylinux_x86_64",
        "cp314-manylinux_x86_64",
        "cp314t-manylinux_x86_64",
    ]
    assert before_alls(build_steps[3]) == [""] * 5
    assert container_engines(build_steps[3]) == [default_container_engine] * 5



import dataclasses
from collections.abc import Mapping, Sequence
from typing import Any, Protocol

import bashlex
import bashlex.errors

from cibuildwheel import bashlex_eval


class EnvironmentParseError(Exception):
    pass


def split_env_items(env_string: str) -> list[str]:
    """Splits space-separated variable assignments into a list of individual assignments.

    >>> split_env_items('VAR=abc')
    ['VAR=abc']
    >>> split_env_items('VAR="a string" THING=3')
    ['VAR="a string"', 'THING=3']
    >>> split_env_items('VAR="a string" THING=\\'single "quotes"\\'')
    ['VAR="a string"', 'THING=\\'single "quotes"\\'']
    >>> split_env_items('VAR="dont \\\\"forget\\\\" about backslashes"')
    ['VAR="dont \\\\"forget\\\\" about backslashes"']
    >>> split_env_items('PATH="$PATH;/opt/cibw_test_path"')
    ['PATH="$PATH;/opt/cibw_test_path"']
    >>> split_env_items('PATH2="something with spaces"')
    ['PATH2="something with spaces"']
    """
    if not env_string:
        return []

    try:
        command_node = bashlex.parsesingle(env_string)
    except bashlex.errors.ParsingError as e:
        raise EnvironmentParseError(env_string) from e

    result = []

    for word_node in command_node.parts:
        part_string = env_string[word_node.pos[0] : word_node.pos[1]]
        result.append(part_string)

    return result


class EnvironmentAssignment(Protocol):
    name: str

    def evaluated_value(
        self,
        *,
        environment: Mapping[str, str],
        executor: bashlex_eval.EnvironmentExecutor | None = None,
    ) -> str:
        """Returns the value of this assignment, as evaluated in the environment"""


class EnvironmentAssignmentRaw:
    """
    An environment variable - a simple name/value pair
    """

    def __init__(self, name: str, value: str):
        self.name = name
        self.value = value

    def __repr__(self) -> str:
        return f"{self.name}={self.value}"

    def evaluated_value(self, **_: Any) -> str:
        return self.value


@dataclasses.dataclass
class EnvironmentAssignmentBash:
    """
    An environment variable, in bash syntax. The value can use bash constructs
    like "$OTHER_VAR" and "$(command arg1 arg2)".
    """

    name: str
    value: str

    def __init__(self, assignment: str):
        name, equals, value = assignment.partition("=")
        if not equals:
            raise EnvironmentParseError(assignment)
        self.name = name
        self.value = value

    def evaluated_value(
        self,
        environment: Mapping[str, str],
        executor: bashlex_eval.EnvironmentExecutor | None = None,
    ) -> str:
        return bashlex_eval.evaluate(self.value, environment=environment, executor=executor)

    def __repr__(self) -> str:
        return f"{self.name}={self.value}"


@dataclasses.dataclass(kw_only=True)
class ParsedEnvironment:
    assignments: list[EnvironmentAssignment]

    def __init__(self, assignments: Sequence[EnvironmentAssignment]) -> None:
        self.assignments = list(assignments)

    def as_dictionary(
        self,
        prev_environment: Mapping[str, str],
        executor: bashlex_eval.EnvironmentExecutor | None = None,
    ) -> dict[str, str]:
        environment = {**prev_environment}

        for assignment in self.assignments:
            value = assignment.evaluated_value(environment=environment, executor=executor)
            environment[assignment.name] = value

        return environment

    def add(self, name: str, value: str, prepend: bool = False) -> None:
        assignment = EnvironmentAssignmentRaw(name=name, value=value)
        if prepend:
            self.assignments.insert(0, assignment)
        else:
            self.assignments.append(assignment)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({[repr(a) for a in self.assignments]!r})"

    def options_summary(self) -> list[EnvironmentAssignment]:
        return self.assignments


def parse_environment(env_string: str) -> ParsedEnvironment:
    env_items = split_env_items(env_string)
    assignments = [EnvironmentAssignmentBash(item) for item in env_items]
    return ParsedEnvironment(assignments=assignments)



import dataclasses
import functools
import inspect
import os
import platform
import re
import shutil
import subprocess
import sys
import typing
from collections.abc import Set
from pathlib import Path
from typing import Literal, assert_never

from filelock import FileLock
from packaging.version import Version

from cibuildwheel import errors
from cibuildwheel.architecture import Architecture
from cibuildwheel.ci import detect_ci_provider
from cibuildwheel.environment import ParsedEnvironment
from cibuildwheel.frontend import BuildFrontendName, get_build_frontend_extra_flags
from cibuildwheel.logger import log
from cibuildwheel.options import Options
from cibuildwheel.selector import BuildSelector
from cibuildwheel.util import resources
from cibuildwheel.util.cmd import call, shell
from cibuildwheel.util.file import CIBW_CACHE_PATH, copy_test_sources, download, move_file
from cibuildwheel.util.helpers import prepare_command, unwrap
from cibuildwheel.util.packaging import find_compatible_wheel, get_pip_version
from cibuildwheel.venv import constraint_flags, find_uv, virtualenv


@functools.cache
def get_macos_version() -> tuple[int, int]:
    """
    Returns the macOS major/minor version, as a tuple, e.g. (10, 15) or (11, 0)

    These tuples can be used in comparisons, e.g.
        (10, 14) <= (11, 0) == True
        (10, 14) <= (10, 16) == True
        (11, 2) <= (11, 0) != True
    """
    version_str, _, _ = platform.mac_ver()
    version = tuple(map(int, version_str.split(".")[:2]))
    if (10, 15) < version < (11, 0):
        # When built against an older macOS SDK, Python will report macOS 10.16
        # instead of the real version.
        version_str = call(
            sys.executable,
            "-sS",
            "-c",
            "import platform; print(platform.mac_ver()[0])",
            env={"SYSTEM_VERSION_COMPAT": "0"},
            capture_stdout=True,
        )
        version = tuple(map(int, version_str.split(".")[:2]))
    return typing.cast("tuple[int, int]", version)


@functools.cache
def get_test_macosx_deployment_target() -> str:
    version = get_macos_version()
    if version >= (11, 0):
        return f"{version[0]}.0"
    return f"{version[0]}.{version[1]}"


def get_macos_sdks() -> list[str]:
    output = call("xcodebuild", "-showsdks", capture_stdout=True)
    return [m.group(1) for m in re.finditer(r"-sdk (macosx\S+)", output)]


@dataclasses.dataclass(frozen=True, kw_only=True)
class PythonConfiguration:
    version: str
    identifier: str
    url: str


def all_python_configurations() -> list[PythonConfiguration]:
    config_dicts = resources.read_python_configs("macos")
    return [PythonConfiguration(**item) for item in config_dicts]


def get_python_configurations(
    build_selector: BuildSelector, architectures: Set[Architecture]
) -> list[PythonConfiguration]:
    python_configurations = all_python_configurations()

    # filter out configs that don't match any of the selected architectures
    python_configurations = [
        c
        for c in python_configurations
        if any(c.identifier.endswith(a.value) for a in architectures)
    ]

    # skip builds as required by BUILD/SKIP
    python_configurations = [c for c in python_configurations if build_selector(c.identifier)]

    # filter-out some cross-compilation configs with PyPy and GraalPy:
    # can't build arm64 on x86_64
    # rosetta allows to build x86_64 on arm64
    if platform.machine() == "x86_64":
        python_configurations_before = set(python_configurations)
        python_configurations = [
            c
            for c in python_configurations
            if not (c.identifier.startswith(("pp", "gp")) and c.identifier.endswith("arm64"))
        ]
        removed_elements = python_configurations_before - set(python_configurations)
        if removed_elements:
            ids = ", ".join(c.identifier for c in removed_elements)
            log.quiet(
                unwrap(
                    f"""
                    Note: {ids}  {"was" if len(removed_elements) == 1 else "were"}
                    selected, but can't be built on x86_64 so will be skipped automatically.
                    """
                )
            )

    return python_configurations


def install_cpython(_tmp: Path, version: str, url: str, free_threading: bool) -> Path:
    ft = "T" if free_threading else ""
    installation_path = Path(f"/Library/Frameworks/Python{ft}.framework/Versions/{version}")
    with FileLock(CIBW_CACHE_PATH / f"cpython{version}.lock"):
        installed_system_packages = call("pkgutil", "--pkgs", capture_stdout=True).splitlines()
        # if this version of python isn't installed, get it from python.org and install
        python_package_identifier = f"org.python.Python.Python{ft}Framework-{version}"
        if python_package_identifier not in installed_system_packages:
            if detect_ci_provider() is None:
                # if running locally, we don't want to install CPython with sudo
                # let the user know & provide a link to the installer
                msg = inspect.cleandoc(
                    f"""
                        Framework CPython {version} not detected as installed in:
                            {installation_path}
                        cibuildwheel will not perform system-wide installs when running outside of CI.
                        To build locally, install CPython {version} on this machine, or, disable this
                        version of Python using CIBW_SKIP=cp{version.replace(".", "")}-macosx_*
                        For portable builds, cibuildwheel needs the official builds from python.org.
                        Download link: {url}
                    """
                )
                raise errors.FatalError(msg)
            python_filename = url.rsplit("/", maxsplit=1)[-1]
            pkg_path = CIBW_CACHE_PATH / "cpython-installer" / python_filename
            if not pkg_path.exists():
                download(url, pkg_path)
            args = []
            if version.startswith("3.13"):
                # Python 3.13 is the first version to have a free-threading option
                args += ["-applyChoiceChangesXML", str(resources.FREE_THREAD_ENABLE_313.resolve())]
            elif version.startswith("3.14"):
                args += ["-applyChoiceChangesXML", str(resources.FREE_THREAD_ENABLE_314.resolve())]
            call("sudo", "installer", "-pkg", pkg_path, *args, "-target", "/")
            pkg_path.unlink()
            env = os.environ.copy()
            env["PIP_DISABLE_PIP_VERSION_CHECK"] = "1"

            if free_threading:
                call(installation_path / f"bin/python{version}t", "-m", "ensurepip", env=env)
                call(
                    installation_path / f"bin/python{version}t",
                    resources.INSTALL_CERTIFI_SCRIPT,
                    env=env,
                )
            else:
                call(installation_path / "bin/python3", resources.INSTALL_CERTIFI_SCRIPT, env=env)

    return installation_path / "bin" / (f"python{version}t" if free_threading else "python3")


def install_pypy(tmp: Path, url: str) -> Path:
    pypy_tar_bz2 = url.rsplit("/", 1)[-1]
    extension = ".tar.bz2"
    assert pypy_tar_bz2.endswith(extension)
    installation_path = CIBW_CACHE_PATH / pypy_tar_bz2[: -len(extension)]
    with FileLock(str(installation_path) + ".lock"):
        if not installation_path.exists():
            downloaded_tar_bz2 = tmp / pypy_tar_bz2
            download(url, downloaded_tar_bz2)
            installation_path.parent.mkdir(parents=True, exist_ok=True)
            call("tar", "-C", installation_path.parent, "-xf", downloaded_tar_bz2)
            downloaded_tar_bz2.unlink()
    return installation_path / "bin" / "pypy3"


def install_graalpy(tmp: Path, url: str) -> Path:
    graalpy_archive = url.rsplit("/", 1)[-1]
    extension = ".tar.gz"
    assert graalpy_archive.endswith(extension)
    installation_path = CIBW_CACHE_PATH / graalpy_archive[: -len(extension)]
    with FileLock(str(installation_path) + ".lock"):
        if not installation_path.exists():
            downloaded_archive = tmp / graalpy_archive
            download(url, downloaded_archive)
            installation_path.mkdir(parents=True)
            # GraalPy top-folder name is inconsistent with archive name
            call("tar", "-C", installation_path, "--strip-components=1", "-xzf", downloaded_archive)
            downloaded_archive.unlink()
    return installation_path / "bin" / "graalpy"


def setup_python(
    tmp: Path,
    python_configuration: PythonConfiguration,
    dependency_constraint: Path | None,
    environment: ParsedEnvironment,
    build_frontend: BuildFrontendName,
) -> tuple[Path, dict[str, str]]:
    uv_path = find_uv()
    use_uv = build_frontend in {"build[uv]", "uv"}

    tmp.mkdir()
    implementation_id = python_configuration.identifier.split("-")[0]
    log.step(f"Installing Python {implementation_id}...")
    if implementation_id.startswith("cp"):
        free_threading = "t-macos" in python_configuration.identifier
        base_python = install_cpython(
            tmp, python_configuration.version, python_configuration.url, free_threading
        )

    elif implementation_id.startswith("pp"):
        base_python = install_pypy(tmp, python_configuration.url)
    elif implementation_id.startswith("gp"):
        base_python = install_graalpy(tmp, python_configuration.url)
    else:
        msg = "Unknown Python implementation"
        raise ValueError(msg)
    assert base_python.exists(), (
        f"{base_python.name} not found, has {list(base_python.parent.iterdir())}"
    )

    log.step("Setting up build environment...")
    venv_path = tmp / "venv"
    env = virtualenv(
        python_configuration.version,
        base_python,
        venv_path,
        dependency_constraint,
        use_uv=use_uv,
    )
    venv_bin_path = venv_path / "bin"
    assert venv_bin_path.exists()
    # Fix issue with site.py setting the wrong `sys.prefix`, `sys.exec_prefix`,
    # `sys.path`, ... for PyPy: https://foss.heptapod.net/pypy/pypy/issues/3175
    # Also fix an issue with the shebang of installed scripts inside the
    # testing virtualenv- see https://github.com/theacodes/nox/issues/44 and
    # https://github.com/pypa/virtualenv/issues/620
    # Also see https://github.com/python/cpython/pull/9516
    env.pop("__PYVENV_LAUNCHER__", None)
    # uv uses this over the current environment's python, so remove it to avoid confusion
    env.pop("PYTHON_VERSION", None)
    env.pop("PYTHON_ARCH", None)
    env.pop("UV_PYTHON", None)

    # we version pip ourselves, so we don't care about pip version checking
    env["PIP_DISABLE_PIP_VERSION_CHECK"] = "1"

    # Apply our environment after pip is ready
    env = environment.as_dictionary(prev_environment=env)

    # check what Python version we're on
    which_python = call("which", "python", env=env, capture_stdout=True).strip()
    print(which_python)
    if which_python != str(venv_bin_path / "python"):
        msg = "python available on PATH doesn't match our installed instance. If you have modified PATH, ensure that you don't overwrite cibuildwheel's entry or insert python above it."
        raise errors.FatalError(msg)
    call("python", "--version", env=env)

    # check what pip version we're on
    if not use_uv:
        assert (venv_bin_path / "pip").exists()
        which_pip = call("which", "pip", env=env, capture_stdout=True).strip()
        print(which_pip)
        if which_pip != str(venv_bin_path / "pip"):
            msg = "pip available on PATH doesn't match our installed instance. If you have modified PATH, ensure that you don't overwrite cibuildwheel's entry or insert pip above it."
            raise errors.FatalError(msg)
        call("pip", "--version", env=env)

    config_is_arm64 = python_configuration.identifier.endswith("arm64")
    config_is_universal2 = python_configuration.identifier.endswith("universal2")

    # Set MACOSX_DEPLOYMENT_TARGET, if the user didn't set it.
    # For arm64, the minimal deployment target is 11.0.
    # On x86_64 (or universal2), use 10.9 as a default.
    # CPython 3.12.6+ needs 10.13.
    # CPython 3.14.0 needs 10.15.
    if config_is_arm64:
        default_target = "11.0"
    elif Version(python_configuration.version) >= Version("3.14"):
        default_target = "10.15"
    elif Version(python_configuration.version) >= Version("3.12"):
        default_target = "10.13"
    elif python_configuration.identifier.startswith("pp") and Version(
        python_configuration.version
    ) >= Version("3.9"):
        default_target = "10.15"
    else:
        default_target = "10.9"
    env.setdefault("MACOSX_DEPLOYMENT_TARGET", default_target)

    # This is a floor, it can't be set lower than the default_target.
    if Version(env["MACOSX_DEPLOYMENT_TARGET"]) < Version(default_target):
        log.warning(
            f"Bumping MACOSX_DEPLOYMENT_TARGET ({env['MACOSX_DEPLOYMENT_TARGET']}) to the minimum required ({default_target})."
        )
        env["MACOSX_DEPLOYMENT_TARGET"] = default_target

    if config_is_arm64:
        # macOS 11 is the first OS with arm64 support, so the wheels
        # have that as a minimum.
        env.setdefault("_PYTHON_HOST_PLATFORM", "macosx-11.0-arm64")
        env.setdefault("ARCHFLAGS", "-arch arm64")
    elif config_is_universal2:
        env.setdefault("_PYTHON_HOST_PLATFORM", "macosx-10.9-universal2")
        env.setdefault("ARCHFLAGS", "-arch arm64 -arch x86_64")
    elif python_configuration.identifier.endswith("x86_64"):
        # even on the macos11.0 Python installer, on the x86_64 side it's
        # compatible back to 10.9.
        env.setdefault("_PYTHON_HOST_PLATFORM", "macosx-10.9-x86_64")
        env.setdefault("ARCHFLAGS", "-arch x86_64")

    building_arm64 = config_is_arm64 or config_is_universal2
    if building_arm64 and get_macos_version() < (10, 16) and "SDKROOT" not in env:
        # xcode 12.2 or higher can build arm64 on macos 10.15 or below, but
        # needs the correct SDK selected.
        sdks = get_macos_sdks()

        # Different versions of Xcode contain different SDK versions...
        # we're happy with anything newer than macOS 11.0
        arm64_compatible_sdks = [s for s in sdks if not s.startswith("macosx10.")]

        if not arm64_compatible_sdks:
            log.warning(
                unwrap(
                    """
                    SDK for building arm64-compatible wheels not found. You need Xcode 12.2 or later
                    to build universal2 or arm64 wheels.
                    """
                )
            )
        else:
            env.setdefault("SDKROOT", arm64_compatible_sdks[0])

    log.step("Installing build tools...")
    match build_frontend:
        case "pip":
            call(
                "pip",
                "install",
                "--upgrade",
                "delocate",
                *constraint_flags(dependency_constraint),
                env=env,
            )
        case "build":
            call(
                "pip",
                "install",
                "--upgrade",
                "delocate",
                "build[virtualenv]",
                *constraint_flags(dependency_constraint),
                env=env,
            )
        case "build[uv]":
            assert uv_path is not None
            call(
                uv_path,
                "pip",
                "install",
                "--python",
                which_python,
                "--upgrade",
                "delocate",
                "build[virtualenv, uv]",
                *constraint_flags(dependency_constraint),
                env=env,
            )
        case "uv":
            assert uv_path is not None
            call(
                uv_path,
                "pip",
                "install",
                "--python",
                which_python,
                "--upgrade",
                "delocate",
                *constraint_flags(dependency_constraint),
                env=env,
            )
        case _:
            assert_never(build_frontend)

    return base_python, env


def build(options: Options, tmp_path: Path) -> None:
    python_configurations = get_python_configurations(
        options.globals.build_selector, options.globals.architectures
    )

    if not python_configurations:
        return

    try:
        before_all_options_identifier = python_configurations[0].identifier
        before_all_options = options.build_options(before_all_options_identifier)

        if before_all_options.before_all:
            log.step("Running before_all...")
            env = before_all_options.environment.as_dictionary(prev_environment=os.environ)
            env.setdefault("MACOSX_DEPLOYMENT_TARGET", "10.9")
            before_all_prepared = prepare_command(
                before_all_options.before_all, project=".", package=before_all_options.package_dir
            )
            shell(before_all_prepared, env=env)

        built_wheels: list[Path] = []

        for config in python_configurations:
            build_options = options.build_options(config.identifier)
            build_frontend = build_options.build_frontend
            use_uv = build_frontend.name in {"build[uv]", "uv"}
            uv_path = find_uv()
            if use_uv and uv_path is None:
                msg = "uv not found"
                raise AssertionError(msg)
            pip = ["pip"] if not use_uv else [str(uv_path), "pip"]
            log.build_start(config.identifier)

            identifier_tmp_dir = tmp_path / config.identifier
            identifier_tmp_dir.mkdir()
            built_wheel_dir = identifier_tmp_dir / "built_wheel"
            repaired_wheel_dir = identifier_tmp_dir / "repaired_wheel"

            config_is_arm64 = config.identifier.endswith("arm64")
            config_is_universal2 = config.identifier.endswith("universal2")

            constraints_path = build_options.dependency_constraints.get_for_python_version(
                version=config.version, tmp_dir=identifier_tmp_dir
            )

            base_python, env = setup_python(
                identifier_tmp_dir / "build",
                config,
                constraints_path,
                build_options.environment,
                build_frontend.name,
            )
            pip_version = None if use_uv else get_pip_version(env)

            compatible_wheel = find_compatible_wheel(built_wheels, config.identifier)
            if compatible_wheel:
                log.step_end()
                print(
                    f"\nFound previously built wheel {compatible_wheel.name}, that's compatible with {config.identifier}. Skipping build step..."
                )
                repaired_wheel = compatible_wheel
            else:
                if build_options.before_build:
                    log.step("Running before_build...")
                    before_build_prepared = prepare_command(
                        build_options.before_build, project=".", package=build_options.package_dir
                    )
                    shell(before_build_prepared, env=env)

                log.step("Building wheel...")
                built_wheel_dir.mkdir()

                extra_flags = get_build_frontend_extra_flags(
                    build_frontend,
                    build_options.build_verbosity,
                    build_options.config_settings,
                    py38=config.identifier[1:].startswith("p38"),
                )

                build_env = env.copy()

                match build_frontend.name:
                    case "pip":
                        # Path.resolve() is needed. Without it pip wheel may try to fetch package from pypi.org
                        # see https://github.com/pypa/cibuildwheel/pull/369
                        call(
                            "python",
                            "-m",
                            "pip",
                            "wheel",
                            build_options.package_dir.resolve(),
                            f"--wheel-dir={built_wheel_dir}",
                            "--no-deps",
                            *extra_flags,
                            env=build_env,
                        )
                    case "build" | "build[uv]":
                        if (
                            use_uv
                            and "--no-isolation" not in extra_flags
                            and "-n" not in extra_flags
                        ):
                            extra_flags.append("--installer=uv")
                        call(
                            "python",
                            "-m",
                            "build",
                            build_options.package_dir,
                            "--wheel",
                            f"--outdir={built_wheel_dir}",
                            *extra_flags,
                            env=build_env,
                        )
                    case "uv":
                        assert uv_path is not None
                        call(
                            uv_path,
                            "build",
                            f"--python={base_python}",
                            build_options.package_dir,
                            "--wheel",
                            f"--out-dir={built_wheel_dir}",
                            *extra_flags,
                            env=build_env,
                        )
                    case _:
                        assert_never(build_frontend)

                built_wheel = next(built_wheel_dir.glob("*.whl"))

                repaired_wheel_dir.mkdir()

                if built_wheel.name.endswith("none-any.whl"):
                    raise errors.NonPlatformWheelError()

                if build_options.repair_command:
                    log.step("Repairing wheel...")

                    if config_is_universal2:
                        delocate_archs = "x86_64,arm64"
                    elif config_is_arm64:
                        delocate_archs = "arm64"
                    else:
                        delocate_archs = "x86_64"

                    repair_command_prepared = prepare_command(
                        build_options.repair_command,
                        wheel=built_wheel,
                        dest_dir=repaired_wheel_dir,
                        delocate_archs=delocate_archs,
                        package=build_options.package_dir,
                        project=".",
                    )
                    shell(repair_command_prepared, env=env)
                else:
                    shutil.move(str(built_wheel), repaired_wheel_dir)

                try:
                    repaired_wheel = next(repaired_wheel_dir.glob("*.whl"))
                except StopIteration:
                    raise errors.RepairStepProducedNoWheelError() from None

                if repaired_wheel.name in {wheel.name for wheel in built_wheels}:
                    raise errors.AlreadyBuiltWheelError(repaired_wheel.name)

                log.step_end()

            if build_options.test_command and build_options.test_selector(config.identifier):
                machine_arch = platform.machine()
                python_arch = call(
                    "python",
                    "-sSc",
                    "import platform; print(platform.machine())",
                    env=env,
                    capture_stdout=True,
                ).strip()
                testing_archs: list[Literal["x86_64", "arm64"]]

                if config_is_arm64:
                    testing_archs = ["arm64"]
                elif config_is_universal2:
                    testing_archs = ["x86_64", "arm64"]
                else:
                    testing_archs = ["x86_64"]

                for testing_arch in testing_archs:
                    if config_is_universal2:
                        arch_specific_identifier = f"{config.identifier}:{testing_arch}"
                        if not build_options.test_selector(arch_specific_identifier):
                            continue

                    if machine_arch == "x86_64" and testing_arch == "arm64":
                        if config_is_arm64:
                            log.warning(
                                unwrap(
                                    """
                                    While arm64 wheels can be built on x86_64, they cannot be
                                    tested. Consider building arm64 wheels natively, if your CI
                                    provider offers this. To silence this warning, set
                                    `CIBW_TEST_SKIP: "*-macosx_arm64"`.
                                    """
                                )
                            )
                        elif config_is_universal2:
                            log.warning(
                                unwrap(
                                    """
                                    While universal2 wheels can be built on x86_64, the arm64 part
                                    of the wheel cannot be tested on x86_64. Consider building
                                    universal2 wheels on an arm64 runner, if your CI provider offers
                                    this. Notably, an arm64 runner can also test the x86_64 part of
                                    the wheel, through Rosetta emulation. To silence this warning,
                                    set `CIBW_TEST_SKIP: "*-macosx_universal2:arm64"`.
                                    """
                                )
                            )
                        else:
                            msg = "unreachable"
                            raise RuntimeError(msg)

                        # skip this test
                        continue

                    is_cp38 = config.identifier.startswith("cp38-")
                    if testing_arch == "arm64" and is_cp38 and python_arch != "arm64":
                        log.warning(
                            unwrap(
                                """
                                While cibuildwheel can build CPython 3.8 universal2/arm64 wheels, we
                                cannot test the arm64 part of them, even when running on an Apple
                                Silicon machine. This is because we use the x86_64 installer of
                                CPython 3.8. See the discussion in
                                https://github.com/pypa/cibuildwheel/pull/1169 for the details. To
                                silence this warning, set `CIBW_TEST_SKIP: "cp38-macosx_*:arm64"`.
                                """
                            )
                        )

                        # skip this test
                        continue

                    log.step(
                        "Testing wheel..."
                        if testing_arch == machine_arch
                        else f"Testing wheel on {testing_arch}..."
                    )

                    arch_prefix = []
                    uv_arch_args = []
                    if testing_arch != machine_arch:
                        if machine_arch == "arm64" and testing_arch == "x86_64":
                            # rosetta2 will provide the emulation with just the arch prefix.
                            arch_prefix = ["arch", "-x86_64"]
                            uv_arch_args = ["--python-platform", "x86_64-apple-darwin"]
                        else:
                            msg = f"don't know how to emulate {testing_arch} on {machine_arch}"
                            raise RuntimeError(msg)

                    # define a custom 'call' function that adds the arch prefix each time
                    call_with_arch = functools.partial(call, *arch_prefix)
                    shell_with_arch = functools.partial(call, *arch_prefix, "/bin/sh", "-c")

                    # set up a virtual environment to install and test from, to make sure
                    # there are no dependencies that were pulled in at build time.
                    venv_dir = identifier_tmp_dir / f"venv-test-{testing_arch}"
                    virtualenv_env = virtualenv(
                        config.version,
                        base_python,
                        venv_dir,
                        None,
                        use_uv=use_uv,
                        env=env,
                        pip_version=pip_version,
                    )
                    if use_uv:
                        pip_install = functools.partial(call, *pip, "install", *uv_arch_args)
                    else:
                        pip_install = functools.partial(call_with_arch, *pip, "install")

                    virtualenv_env["MACOSX_DEPLOYMENT_TARGET"] = get_test_macosx_deployment_target()

                    virtualenv_env = build_options.test_environment.as_dictionary(
                        prev_environment=virtualenv_env
                    )

                    # check that we are using the Python from the virtual environment
                    call_with_arch("which", "python", env=virtualenv_env)

                    if build_options.before_test:
                        before_test_prepared = prepare_command(
                            build_options.before_test,
                            project=".",
                            package=build_options.package_dir,
                        )
                        shell_with_arch(before_test_prepared, env=virtualenv_env)

                    # install the wheel
                    if is_cp38 and python_arch == "x86_64":
                        virtualenv_env_install_wheel = virtualenv_env.copy()
                        virtualenv_env_install_wheel["SYSTEM_VERSION_COMPAT"] = "0"
                        log.notice(
                            unwrap(
                                """
                                Setting SYSTEM_VERSION_COMPAT=0 to ensure CPython 3.8 can get
                                correct macOS version and allow installation of wheels with
                                MACOSX_DEPLOYMENT_TARGET >= 11.0.
                                See https://github.com/pypa/cibuildwheel/issues/1767 for the
                                details.
                                """
                            )
                        )
                    else:
                        virtualenv_env_install_wheel = virtualenv_env

                    pip_install(
                        f"{repaired_wheel}{build_options.test_extras}",
                        env=virtualenv_env_install_wheel,
                    )

                    # test the wheel
                    if build_options.test_requires:
                        pip_install(
                            *build_options.test_requires,
                            env=virtualenv_env_install_wheel,
                        )

                    # run the tests from a temp dir, with an absolute path in the command
                    # (this ensures that Python runs the tests against the installed wheel
                    # and not the repo code)
                    test_command_prepared = prepare_command(
                        build_options.test_command,
                        project=Path.cwd(),
                        package=build_options.package_dir.resolve(),
                        wheel=repaired_wheel,
                    )

                    test_cwd = identifier_tmp_dir / "test_cwd"

                    if build_options.test_sources:
                        # only create test_cwd if it doesn't already exist - it
                        # may have been created during a previous `testing_arch`
                        if not test_cwd.exists():
                            test_cwd.mkdir()
                            copy_test_sources(
                                build_options.test_sources,
                                Path.cwd(),
                                test_cwd,
                            )
                    else:
                        # Use the test_fail.py file to raise a nice error if the user
                        # tries to run tests in the cwd
                        test_cwd.mkdir(exist_ok=True)
                        (test_cwd / "test_fail.py").write_text(
                            resources.TEST_FAIL_CWD_FILE.read_text()
                        )

                    shell_with_arch(test_command_prepared, cwd=test_cwd, env=virtualenv_env)

            # we're all done here; move it to output (overwrite existing)
            output_wheel = None
            if compatible_wheel is None:
                output_wheel = build_options.output_dir.joinpath(repaired_wheel.name)
                moved_wheel = move_file(repaired_wheel, output_wheel)
                if moved_wheel != output_wheel.resolve():
                    log.warning(
                        f"{repaired_wheel} was moved to {moved_wheel} instead of {output_wheel}"
                    )
                built_wheels.append(output_wheel)

            # clean up
            shutil.rmtree(identifier_tmp_dir)

            log.build_end(output_wheel)
    except subprocess.CalledProcessError as error:
        msg = f"Command {error.cmd} failed with code {error.returncode}. {error.stdout or ''}"
        raise errors.FatalError(msg) from error



import os
import platform
import shutil
import subprocess
import textwrap
from pathlib import Path

import pytest

from cibuildwheel.ci import CIProvider, detect_ci_provider

from . import test_projects, utils

pytestmark = pytest.mark.ios

CIBW_PLATFORM = os.environ.get("CIBW_PLATFORM", "ios")
if CIBW_PLATFORM != "ios":
    pytest.skip(f"{CIBW_PLATFORM=}", allow_module_level=True)

basic_project_files = {
    "tests/test_platform.py": f"""
import platform
from unittest import TestCase

class TestPlatform(TestCase):
    def test_platform(self) -> None:
        self.assertEqual(platform.machine(), "{platform.machine()}")

"""
}


def skip_if_ios_testing_not_supported() -> None:
    """Skip the test if iOS testing is not supported on this machine."""
    if utils.get_platform() != "macos":
        pytest.skip("this test can only run on macOS")
    if utils.get_xcode_version() < (13, 0):
        pytest.skip("this test only works with Xcode 13.0 or greater")
    if detect_ci_provider() == CIProvider.cirrus_ci:
        pytest.skip(
            "iOS testing not currently supported on Cirrus CI due to a failure "
            "to start the simulator."
        )


# iOS tests shouldn't be run in parallel, because they're dependent on calling
# Xcode, and starting a simulator. These are both multi-threaded operations, and
# it's easy to overload the CI machine if there are multiple test processes
# running multithreaded processes. Therefore, they're put in the serial group,
# which is guaranteed to run single-process.
# This can also fail the first time sometimes.
@pytest.mark.serial
@pytest.mark.flaky(reruns=2)
@pytest.mark.parametrize(
    "build_config",
    [
        # Default to the pip build frontend
        {"CIBW_PLATFORM": "ios"},
        # Also check the build frontend
        {"CIBW_PLATFORM": "ios", "CIBW_BUILD_FRONTEND": "build"},
    ],
)
def test_ios_platforms(
    tmp_path: Path,
    build_config: dict[str, str],
    monkeypatch: pytest.MonkeyPatch,
    capfd: pytest.CaptureFixture[str],
) -> None:
    skip_if_ios_testing_not_supported()

    # Create a temporary "bin" directory, symlink a tool that we know eixsts
    # (/usr/bin/true) into that location under a name that should be unique,
    # and add the temp bin directory to the PATH.
    tools_dir = tmp_path / "bin"
    tools_dir.mkdir()
    true_app = shutil.which("true")
    assert true_app is not None
    tools_dir.joinpath("does-exist").symlink_to(true_app)

    monkeypatch.setenv("PATH", str(tools_dir), prepend=os.pathsep)

    # Generate a test project that has an additional before-build step using the
    # known-to-exist tool.
    project_dir = tmp_path / "project"
    setup_py_add = "import subprocess\nsubprocess.run('does-exist', check=True)\n"
    basic_project = test_projects.new_c_project(setup_py_add=setup_py_add)
    basic_project.files.update(basic_project_files)
    basic_project.generate(project_dir)

    # Build and test the wheels. Mark the "does-exist" tool as a cross-build
    # tool, and invoke it during a `before-build` step. It will also be invoked
    # when `setup.py` is invoked.
    #
    # Tests are only executed on simulator. The test suite passes if it's
    # running on the same architecture as the current platform.
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_BEFORE_BUILD": "does-exist",
            "CIBW_XBUILD_TOOLS": "does-exist",
            "CIBW_TEST_SOURCES": "tests",
            "CIBW_TEST_COMMAND": "python -m this && python -m unittest discover tests test_platform.py",
            "CIBW_BUILD_VERBOSITY": "1",
            **build_config,
        },
    )

    # The expected wheels were produced.
    expected_wheels = utils.expected_wheels("spam", "0.1.0", platform="ios")
    assert set(actual_wheels) == set(expected_wheels)

    # The user was notified that the cross-build tool was found.
    captured = capfd.readouterr()
    assert "'does-exist' will be included in the cross-build environment" in captured.out

    # Make sure the first command ran
    assert "Zen of Python" in captured.out


@pytest.mark.serial
def test_no_test_sources(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    """Build will provide a helpful error if pytest is run and test-sources is not defined."""
    skip_if_ios_testing_not_supported()

    project_dir = tmp_path / "project"
    basic_project = test_projects.new_c_project()
    basic_project.files.update(basic_project_files)
    basic_project.generate(project_dir)

    with pytest.raises(subprocess.CalledProcessError):
        utils.cibuildwheel_run(
            project_dir,
            add_env={
                "CIBW_PLATFORM": "ios",
                "CIBW_TEST_REQUIRES": "pytest",
                "CIBW_TEST_COMMAND": "python -m pytest",
                "CIBW_XBUILD_TOOLS": "",
            },
        )

    # The error message indicates the configuration issue.
    captured = capfd.readouterr()
    assert (
        "you must copy your test files to the testbed app by setting the `test-sources` option"
        in captured.out + captured.err
    )


def test_ios_testing_with_placeholder(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    """
    Tests with the {project} placeholder are not supported on iOS, because the test command
    is run in the simulator.
    """
    skip_if_ios_testing_not_supported()

    project_dir = tmp_path / "project"
    basic_project = test_projects.new_c_project()
    basic_project.files.update(basic_project_files)
    basic_project.generate(project_dir)

    with pytest.raises(subprocess.CalledProcessError):
        utils.cibuildwheel_run(
            project_dir,
            add_env={
                "CIBW_PLATFORM": "ios",
                "CIBW_TEST_REQUIRES": "pytest",
                "CIBW_TEST_COMMAND": "pytest {project}/tests",
                "CIBW_XBUILD_TOOLS": "",
            },
        )

    # The error message indicates the configuration issue.
    captured = capfd.readouterr()
    assert "iOS tests cannot use placeholders" in captured.out + captured.err


@pytest.mark.serial
@pytest.mark.flaky(reruns=2)
def test_ios_test_command_short_circuit(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    skip_if_ios_testing_not_supported()

    project_dir = tmp_path / "project"
    basic_project = test_projects.new_c_project()
    basic_project.files.update(basic_project_files)
    basic_project.generate(project_dir)

    with pytest.raises(subprocess.CalledProcessError):
        # `python -m not_a_module` will fail, so `python -m this` should not be run.
        utils.cibuildwheel_run(
            project_dir,
            add_env={
                "CIBW_PLATFORM": "ios",
                "CIBW_XBUILD_TOOLS": "",
                "CIBW_TEST_SOURCES": "tests",
                "CIBW_TEST_COMMAND": "python -m not_a_module && python -m this",
                "CIBW_BUILD_VERBOSITY": "1",
            },
        )

    captured = capfd.readouterr()

    assert "No module named not_a_module" in captured.out + captured.err
    # assert that `python -m this` was not run
    assert "Zen of Python" not in captured.out + captured.err


def test_missing_xbuild_tool(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    """Build will fail if xbuild-tools references a non-existent tool."""
    skip_if_ios_testing_not_supported()

    project_dir = tmp_path / "project"
    basic_project = test_projects.new_c_project()
    basic_project.files.update(basic_project_files)
    basic_project.generate(project_dir)

    with pytest.raises(subprocess.CalledProcessError):
        utils.cibuildwheel_run(
            project_dir,
            add_env={
                "CIBW_PLATFORM": "ios",
                "CIBW_TEST_COMMAND": "python -m tests",
                "CIBW_XBUILD_TOOLS": "does-not-exist",
            },
        )

    # The error message indicates the problem tool.
    captured = capfd.readouterr()
    assert "Could not find a 'does-not-exist' executable on the path." in captured.err


def test_no_xbuild_tool_definition(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    """Build will succeed with a warning if there is no xbuild-tools definition."""
    skip_if_ios_testing_not_supported()

    project_dir = tmp_path / "project"
    basic_project = test_projects.new_c_project()
    basic_project.files.update(basic_project_files)
    basic_project.generate(project_dir)

    # Build, but don't test the wheels; we're only checking that the right
    # warning was raised.
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_PLATFORM": "ios",
            "CIBW_BUILD": "cp313-*",
            "CIBW_TEST_SKIP": "*",
        },
    )

    # The expected wheels were produced.
    expected_wheels = utils.expected_wheels(
        "spam",
        "0.1.0",
        platform="ios",
        python_abi_tags=["cp313-cp313"],
    )
    assert set(actual_wheels) == set(expected_wheels)

    # The user was notified that there was no cross-build tool definition.
    captured = capfd.readouterr()
    assert "Your project configuration does not define any cross-build tools." in captured.err


def test_empty_xbuild_tool_definition(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    """Build will succeed with no warning if there is an empty xbuild-tools definition."""
    skip_if_ios_testing_not_supported()

    project_dir = tmp_path / "project"
    basic_project = test_projects.new_c_project()
    basic_project.files.update(basic_project_files)
    basic_project.generate(project_dir)

    # Build, but don't test the wheels; we're only checking that a warning
    # wasn't raised.
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_PLATFORM": "ios",
            "CIBW_BUILD": "cp313-*",
            "CIBW_TEST_SKIP": "*",
            "CIBW_XBUILD_TOOLS": "",
        },
    )

    expected_wheels = utils.expected_wheels(
        "spam", "0.1.0", platform="ios", python_abi_tags=["cp313-cp313"]
    )
    assert set(actual_wheels) == set(expected_wheels)

    # The warnings about cross-build notifications were silenced.
    captured = capfd.readouterr()
    assert "Your project configuration does not define any cross-build tools." not in captured.err


@pytest.mark.serial
def test_ios_test_command_without_python_dash_m(
    tmp_path: Path, capfd: pytest.CaptureFixture[str]
) -> None:
    """pytest should be able to run without python -m, but it should warn."""
    skip_if_ios_testing_not_supported()

    project_dir = tmp_path / "project"

    project = test_projects.new_c_project()
    project.files["tests/__init__.py"] = ""
    project.files["tests/test_spam.py"] = textwrap.dedent("""
        import spam
        def test_spam() -> None:
            assert spam.filter("spam") == 0
            assert spam.filter("ham") != 0
    """)
    project.generate(project_dir)

    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_PLATFORM": "ios",
            "CIBW_BUILD": "cp313-*",
            "CIBW_TEST_COMMAND": "pytest ./tests",
            "CIBW_TEST_SOURCES": "tests",
            "CIBW_TEST_REQUIRES": "pytest",
            "CIBW_XBUILD_TOOLS": "",
        },
    )

    expected_wheels = utils.expected_wheels(
        "spam", "0.1.0", platform="ios", python_abi_tags=["cp313-cp313"]
    )
    assert set(actual_wheels) == set(expected_wheels)

    _, err = capfd.readouterr()

    assert "iOS tests configured with a test command which doesn't start with 'python -m'" in err


def test_ios_test_command_invalid(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    """Test command should raise an error if it's clearly invalid."""
    skip_if_ios_testing_not_supported()

    project_dir = tmp_path / "project"
    basic_project = test_projects.new_c_project()
    basic_project.files["./my_test_script.sh"] = "echo hello"
    basic_project.generate(project_dir)

    with pytest.raises(subprocess.CalledProcessError):
        utils.cibuildwheel_run(
            project_dir,
            add_env={
                "CIBW_PLATFORM": "ios",
                "CIBW_TEST_COMMAND": "./my_test_script.sh",
                "CIBW_TEST_SOURCES": "./my_test_script.sh",
                "CIBW_XBUILD_TOOLS": "",
            },
        )
    _, err = capfd.readouterr()
    assert "iOS tests configured with a test command which doesn't start with 'python -m'" in err


def test_repair_step(tmp_path: Path) -> None:
    """Build will succeed & the custom repair step is called."""
    skip_if_ios_testing_not_supported()

    project_dir = tmp_path / "project"
    basic_project = test_projects.new_c_project()
    basic_project.files.update(basic_project_files)
    basic_project.generate(project_dir)

    # Build, but don't test the wheels; we're only checking that the right
    # warning was raised.
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_PLATFORM": "ios",
            "CIBW_BUILD": "cp314-*",
            "CIBW_TEST_SKIP": "*",
            "CIBW_REPAIR_WHEEL_COMMAND": "touch .ios-repair-step && mv {wheel} {dest_dir}",
        },
    )

    # The expected wheels were produced.
    expected_wheels = utils.expected_wheels(
        "spam",
        "0.1.0",
        platform="ios",
        python_abi_tags=["cp314-cp314"],
    )
    assert set(actual_wheels) == set(expected_wheels)

    # The custom repair step is called.
    assert project_dir.joinpath(".ios-repair-step").is_file()



import os
import shutil
import ssl
import tarfile
import time
import urllib.request
from collections.abc import Callable
from pathlib import Path, PurePath
from typing import Final
from zipfile import ZipFile

import certifi
from platformdirs import user_cache_path

from cibuildwheel.errors import FatalError

DEFAULT_CIBW_CACHE_PATH: Final[Path] = user_cache_path(appname="cibuildwheel", appauthor="pypa")
CIBW_CACHE_PATH: Final[Path] = Path(
    os.environ.get("CIBW_CACHE_PATH", DEFAULT_CIBW_CACHE_PATH)
).resolve()


def ensure_cache_sentinel(cache_path: Path) -> None:
    """Create a sentinel file to mark a cibuildwheel cache directory.

    We help prevent accidental deletion of non-cache directories with this.

    Args:
        cache_path: The cache directory path where the sentinel should be created
    """
    if cache_path.exists():
        sentinel_file = cache_path / "CACHEDIR.TAG"
        if not sentinel_file.exists():
            sentinel_file.write_text(
                "Signature: 8a477f597d28d172789f06886806bc55\n"
                "# This file is a cache directory tag created by cibuildwheel.\n"
                "# For information about cache directory tags, see:\n"
                "# https://www.brynosaurus.com/cachedir/",
                encoding="utf-8",
            )


def download(url: str, dest: Path) -> None:
    print(f"+ Download {url} to {dest}")
    dest_dir = dest.parent
    dest_dir.mkdir(parents=True, exist_ok=True)

    # we've had issues when relying on the host OS' CA certificates on Windows,
    # so we use certifi (this sounds odd but requests also does this by default)
    cafile = os.environ.get("SSL_CERT_FILE", certifi.where())
    context = ssl.create_default_context(cafile=cafile)
    repeat_num = 3
    for i in range(repeat_num):
        try:
            with urllib.request.urlopen(url, context=context) as response:
                dest.write_bytes(response.read())
                return

        except OSError:
            if i == repeat_num - 1:
                raise
            time.sleep(3)


def extract_zip(zip_src: Path, dest: Path) -> None:
    """Extracts a zip and correctly sets permissions on extracted files.

    Notes:
        - sets permissions to the same values as they were set in the archive
        - files with no clear permissions in `external_attr` will be extracted with default values
    """
    with ZipFile(zip_src) as zip_:
        for zinfo in zip_.filelist:
            zip_.extract(zinfo, dest)

            # Set permissions to the same values as they were set in the archive
            # We have to do this manually due to https://github.com/python/cpython/issues/59999
            permissions = (zinfo.external_attr >> 16) & 0o777
            if permissions != 0:
                dest.joinpath(zinfo.filename).chmod(permissions)


def extract_tar(tar_src: Path, dest: Path) -> None:
    """Extracts a tar file using the stdlib 'tar' filter.

    See: https://docs.python.org/3/library/tarfile.html#tarfile.tar_filter for filter details
    """
    with tarfile.open(tar_src) as tar_:
        tar_.extraction_filter = getattr(tarfile, "tar_filter", (lambda member, _: member))
        tar_.extractall(dest)


def move_file(src_file: Path, dst_file: Path) -> Path:
    """Moves a file safely while avoiding potential semantic confusion:

    1. `dst_file` must point to the target filename, not a directory
    2. `dst_file` will be overwritten if it already exists
    3. any missing parent directories will be created

    Returns the fully resolved Path of the resulting file.

    Raises:
        NotADirectoryError: If any part of the intermediate path to `dst_file` is an existing file
        IsADirectoryError: If `dst_file` points directly to an existing directory
    """
    src_file = src_file.resolve(strict=True)
    dst_file = dst_file.resolve()

    if dst_file.is_dir():
        msg = "dst_file must be a valid target filename, not an existing directory."
        raise IsADirectoryError(msg)
    dst_file.unlink(missing_ok=True)
    dst_file.parent.mkdir(parents=True, exist_ok=True)

    # using shutil.move() as Path.rename() is not guaranteed to work across filesystem boundaries
    # explicit str() needed for Python 3.8
    resulting_file = shutil.move(str(src_file), str(dst_file))
    return Path(resulting_file).resolve(strict=True)


def copy_into_local(src: Path, dst: PurePath) -> None:
    """Copy a path from src to dst, regardless of whether it's a file or a directory."""
    # Ensure the target folder location exists
    Path(dst.parent).mkdir(exist_ok=True, parents=True)

    if src.is_dir():
        shutil.copytree(src, dst)
    else:
        shutil.copy(src, dst)


def copy_test_sources(
    test_sources: list[str],
    project_dir: Path,
    test_dir: PurePath,
    copy_into: Callable[[Path, PurePath], None] = copy_into_local,
) -> None:
    """Copy the list of test sources from the package to the test directory.

    :param test_sources: A list of test paths, relative to the project_dir.
    :param project_dir: The root of the project.
    :param test_dir: The folder where test sources should be placed.
    :param copy_into: The copy function to use. By default, does a local
        filesystem copy; but an OCIContainer.copy_info method (or equivalent)
        can be provided.
    """
    for test_path in test_sources:
        source = project_dir.resolve() / test_path

        if not source.exists():
            msg = f"Test source {test_path} does not exist."
            raise FatalError(msg)

        copy_into(source, test_dir / test_path)



#!/usr/bin/env python3


import os
import re
import shutil
import subprocess
import sys
import textwrap
import time
import typing
from pathlib import Path
from urllib.parse import quote

import click

DIR = Path(__file__).parent.parent.resolve()

BuildBackend = typing.Literal["setuptools", "meson"]


def shell(cmd: str, *, check: bool, **kwargs: object) -> subprocess.CompletedProcess[str]:
    return subprocess.run([cmd], shell=True, check=check, **kwargs)  # type: ignore[call-overload, no-any-return]


def git_repo_has_changes() -> bool:
    unstaged_changes = shell("git diff-index --quiet HEAD --", check=False).returncode != 0
    staged_changes = shell("git diff-index --quiet --cached HEAD --", check=False).returncode != 0
    return unstaged_changes or staged_changes


def generate_project(path: Path, build_backend: BuildBackend) -> None:
    sys.path.insert(0, "")
    match build_backend:
        case "meson":
            from test.test_projects.meson import new_meson_project as new_project  # noqa: PLC0415
        case "setuptools":
            from test.test_projects.setuptools import new_c_project as new_project  # noqa: PLC0415
        case _:
            typing.assert_never(build_backend)

    project = new_project()
    project.generate(path)


class CIService(typing.NamedTuple):
    name: str
    dst_config_path: str
    badge_md: str
    config_file_transform: typing.Callable[[str, str], str] = lambda x, _: x  # identity by default


def github_config_file_transform(content: str, git_ref: str) -> str:
    # one of the the github configs only builds on main, so we need to remove that restriction
    # so our example build will run on the test branch.
    #
    # replace:
    # """
    # push:
    #   branches:
    #     - main
    # """
    # with:
    # """
    # push:
    # """"
    content = re.sub(
        r"push:\n\s+branches:\n\s+- main",
        "push:",
        content,
    )

    # use the version of cibuildwheel from the current commit, not the latest
    # release
    # replace:
    # """
    # uses: pypa/cibuildwheel@v3.3.1
    # """
    # with:
    # """
    # uses: pypa/cibuildwheel@<latest commit hash>
    # """
    content = re.sub(
        r"uses: pypa/cibuildwheel@v.*",
        f"uses: pypa/cibuildwheel@{git_ref}",
        content,
    )
    return content


services = [
    CIService(
        name="azure-pipelines",
        dst_config_path="azure-pipelines.yml",
        badge_md="[![Build Status](https://dev.azure.com/joerick0429/cibuildwheel/_apis/build/status/pypa.cibuildwheel?branchName={branch})](https://dev.azure.com/joerick0429/cibuildwheel/_build/latest?definitionId=2&branchName={branch})",
    ),
    CIService(
        name="circleci",
        dst_config_path=".circleci/config.yml",
        badge_md="[![CircleCI](https://circleci.com/gh/pypa/cibuildwheel/tree/{branch_escaped}.svg?style=svg)](https://circleci.com/gh/pypa/cibuildwheel/tree/{branch})",
    ),
    CIService(
        name="github",
        dst_config_path=".github/workflows/example.yml",
        badge_md="[![Build](https://github.com/pypa/cibuildwheel/actions/workflows/example.yml/badge.svg?branch={branch})](https://github.com/pypa/cibuildwheel/actions?query=branch%3A{branch})",
        config_file_transform=github_config_file_transform,
    ),
    CIService(
        name="travis-ci",
        dst_config_path=".travis.yml",
        badge_md="[![Build Status](https://app.travis-ci.com/pypa/cibuildwheel.svg?branch={branch})](https://app.travis-ci.com/pypa/cibuildwheel)",
    ),
    CIService(
        name="gitlab",
        dst_config_path=".gitlab-ci.yml",
        badge_md="[![Gitlab](https://gitlab.com/joerick/cibuildwheel/badges/{branch}/pipeline.svg)](https://gitlab.com/joerick/cibuildwheel/-/commits/{branch})",
    ),
    CIService(
        name="cirrus-ci",
        dst_config_path=".cirrus.yml",
        badge_md="[![Cirrus CI](https://api.cirrus-ci.com/github/pypa/cibuildwheel.svg?branch={branch})](https://cirrus-ci.com/github/pypa/cibuildwheel/{branch})",
    ),
]


def ci_service_for_config_file(config_file: Path) -> CIService:
    filename = config_file.name
    try:
        return next(s for s in services if filename.startswith(s.name))
    except StopIteration:
        msg = f"unknown ci service for config file {config_file}"
        raise ValueError(msg) from None


@click.command()
@click.argument("config_files", nargs=-1, type=click.Path())
@click.option("--build-backend", type=click.Choice(["setuptools", "meson"]), default="setuptools")
def run_example_ci_configs(
    config_files: list[str], build_backend: BuildBackend = "setuptools"
) -> None:
    """
    Test the example configs. If no files are specified, will test
    examples/*-minimal.yml
    """

    if len(config_files) == 0:
        config_file_paths = list(Path("examples").glob("*-minimal.yml"))
    else:
        config_file_paths = [Path(f) for f in config_files]

    # check each CI service has at most 1 config file
    configs_by_service = set()
    for config_file in config_file_paths:
        service = ci_service_for_config_file(config_file)
        if service.name in configs_by_service:
            msg = "You cannot specify more than one config per CI service"
            raise Exception(msg)
        configs_by_service.add(service.name)

    if git_repo_has_changes():
        print("Your git repo has uncommitted changes. Commit or stash before continuing.")
        sys.exit(1)

    previous_branch = shell(
        "git rev-parse --abbrev-ref HEAD", check=True, capture_output=True, encoding="utf8"
    ).stdout.strip()
    git_ref = shell(
        "git rev-parse HEAD", check=True, capture_output=True, encoding="utf8"
    ).stdout.strip()

    timestamp = time.strftime("%Y-%m-%dT%H-%M-%S", time.gmtime())
    branch_name = f"example-config-test---{previous_branch}-{timestamp}"

    try:
        shell(f"git checkout --orphan {branch_name}", check=True)

        example_project = Path("example_root")
        generate_project(example_project, build_backend=build_backend)

        for config_file in config_file_paths:
            service = ci_service_for_config_file(config_file)
            dst_config_file = example_project / service.dst_config_path

            dst_config_file.parent.mkdir(parents=True, exist_ok=True)

            contents = config_file.read_text(encoding="utf8")
            contents = service.config_file_transform(contents, git_ref)
            dst_config_file.write_text(contents, encoding="utf8")

        subprocess.run(["git", "add", example_project], check=True)
        message = textwrap.dedent(
            f"""\
            Test example CI configs

            Testing files: {[str(f) for f in config_files]}
            Generated from branch: {previous_branch}
            Time: {timestamp}
            """
        )
        subprocess.run(["git", "commit", "--no-verify", "--message", message], check=True)
        shell(f"git subtree --prefix={example_project} push origin {branch_name}", check=True)

        print("---")
        print()
        print("> **Examples test run**")
        print("> ")
        print(f"> Branch: [{branch_name}](https://github.com/pypa/cibuildwheel/tree/{branch_name})")
        print("> ")
        print("> | Service | Config | Status |")
        print("> |---|---|---|")
        for config_file in config_file_paths:
            service = ci_service_for_config_file(config_file)
            badge = service.badge_md.format(
                branch=branch_name, branch_escaped=quote(branch_name, safe="")
            )
            print(f"> | {service.name} | `{config_file}` | {badge} |")
        print("> ")
        print(f"> Generated by `{' '.join(sys.argv)}`")
        print()
        print("---")
    finally:
        # remove any local changes
        shutil.rmtree(example_project, ignore_errors=True)
        shell("git checkout -- .", check=True)
        shell(f"git checkout {previous_branch}", check=True)
        shell(f"git branch -D --force {branch_name}", check=True)


if __name__ == "__main__":
    os.chdir(DIR)
    run_example_ci_configs(standalone_mode=True)



import ast
import configparser
import contextlib
from pathlib import Path
from typing import Any

import dependency_groups


def get_parent(node: ast.AST | None, depth: int = 1) -> ast.AST | None:
    for _ in range(depth):
        node = getattr(node, "parent", None)
    return node


def is_main(parent: ast.AST | None) -> bool:
    match parent:
        case ast.If(test=ast.Compare(left=left, ops=[ast.Eq()], comparators=[comp])):
            values = {left, comp}
            mains = {x for x in values if isinstance(x, ast.Constant) and x.value == "__main__"}
            if len(mains) != 1:
                return False
            consts = {x for x in values if isinstance(x, ast.Name) and x.id == "__name__"}
            return len(consts) == 1
        case _:
            return False


class Analyzer(ast.NodeVisitor):
    def __init__(self) -> None:
        self.requires_python: str | None = None

    def visit(self, node: ast.AST) -> None:
        for inner_node in ast.walk(node):
            for child in ast.iter_child_nodes(inner_node):
                child.parent = inner_node  # type: ignore[attr-defined]
        super().visit(node)

    def visit_keyword(self, node: ast.keyword) -> None:
        # Must not be nested except for if __name__ == "__main__"

        self.generic_visit(node)
        # This will be Module -> Expr -> Call -> keyword
        parent = get_parent(node, 4)
        unnested = parent is None

        # This will be Module -> If -> Expr -> Call -> keyword
        name_main_unnested = (
            parent is not None and get_parent(parent) is None and is_main(get_parent(node, 3))
        )

        match node:
            case ast.keyword(arg="python_requires", value=ast.Constant(value=str() as version)):
                if unnested or name_main_unnested:
                    self.requires_python = version


def setup_py_python_requires(content: str) -> str | None:
    try:
        tree = ast.parse(content)
        analyzer = Analyzer()
        analyzer.visit(tree)
        return analyzer.requires_python or None
    except Exception:  # pylint: disable=broad-except
        return None


def get_requires_python_str(package_dir: Path, pyproject_toml: dict[str, Any] | None) -> str | None:
    """Return the python requires string from the most canonical source available, or None"""

    # Read in from pyproject.toml:project.requires-python
    with contextlib.suppress(KeyError, IndexError, TypeError):
        return str((pyproject_toml or {})["project"]["requires-python"])

    # Read in from setup.cfg:options.python_requires
    config = configparser.ConfigParser()
    with contextlib.suppress(FileNotFoundError):
        config.read(package_dir / "setup.cfg")
        with contextlib.suppress(KeyError, IndexError, TypeError):
            return str(config["options"]["python_requires"])

    setup_py = package_dir / "setup.py"
    with contextlib.suppress(FileNotFoundError), setup_py.open(encoding="utf8") as f2:
        return setup_py_python_requires(f2.read())

    return None


def resolve_dependency_groups(
    pyproject_toml: dict[str, Any] | None, *groups: str
) -> tuple[str, ...]:
    """
    Get the packages in dependency-groups for a package.
    """

    if not groups:
        return ()

    if pyproject_toml is None:
        msg = f"Didn't find a pyproject.toml, so can't read [dependency-groups] {groups!r} from it!"
        raise FileNotFoundError(msg)

    try:
        dependency_groups_toml = pyproject_toml["dependency-groups"]
    except KeyError:
        msg = f"Didn't find [dependency-groups] in pyproject.toml, which is needed to resolve {groups!r}."
        raise KeyError(msg) from None

    return dependency_groups.resolve(dependency_groups_toml, *groups)



import json
import platform
import re
import subprocess
import textwrap
from pathlib import Path

import pytest

from cibuildwheel.util import resources

from . import test_projects, utils

VERSION_REGEX = r"([\w-]+)==([^\s]+)"

CHECK_VERSIONS_SCRIPT = """\
'''
Checks that the versions in the env var EXPECTED_VERSIONS match those
installed in the active venv.
'''
import os, subprocess, sys, json

versions_raw = json.loads(
    subprocess.check_output([
        sys.executable, '-m', 'pip', 'list', '--format=json',
    ], text=True)
)
versions = {item['name']: item['version'] for item in versions_raw}
expected_versions = json.loads(os.environ['EXPECTED_VERSIONS'])

for name, expected_version in expected_versions.items():
    if name not in versions:
        continue
    if versions[name] != expected_version:
        raise SystemExit(f'error: {name} version should equal {expected_version}. Versions: {versions}')
"""


def test_check_versions_script(
    tmp_path: Path, build_frontend_env_nouv: dict[str, str], capfd: pytest.CaptureFixture[str]
) -> None:
    if utils.get_platform() == "linux":
        pytest.skip("we don't test dependency versions on linux, refer to other tests")

    # sanity check that the CHECK_VERSIONS_SCRIPT fails when it should
    project_dir = tmp_path / "project"
    test_projects.new_c_project().generate(project_dir)

    expected_versions = {
        "pip": "0.0.1",
        "build": "0.0.2",
    }
    script = project_dir / "check_versions.py"
    script.write_text(CHECK_VERSIONS_SCRIPT)

    with pytest.raises(subprocess.CalledProcessError):
        utils.cibuildwheel_run(
            project_dir,
            add_env={
                "CIBW_BEFORE_BUILD": f"python {script.name}",
                "EXPECTED_VERSIONS": json.dumps(expected_versions),
                **build_frontend_env_nouv,
            },
        )

    captured = capfd.readouterr()

    assert (
        "error: pip version should equal 0.0.1" in captured.err
        or "error: build version should equal 0.0.2" in captured.err
    )


def get_versions_from_constraint_file(constraint_file: Path) -> dict[str, str]:
    constraint_file_text = constraint_file.read_text(encoding="utf-8")

    return dict(re.findall(VERSION_REGEX, constraint_file_text))


@pytest.mark.parametrize("python_version", ["3.8", "3.12"])
def test_pinned_versions(
    tmp_path: Path, python_version: str, build_frontend_env_nouv: dict[str, str]
) -> None:
    if utils.get_platform() == "linux":
        pytest.skip("linux doesn't pin individual tool versions, it pins manylinux images instead")
    if python_version != "3.12" and utils.get_platform() == "pyodide":
        pytest.skip(f"pyodide does not support Python {python_version}")
    if (
        python_version == "3.8"
        and utils.get_platform() == "windows"
        and platform.machine() == "ARM64"
    ):
        pytest.skip(f"Windows ARM64 does not support Python {python_version}")

    project_dir = tmp_path / "project"
    test_projects.new_c_project().generate(project_dir)

    # read the expected versions from the appropriate constraint file
    version_no_dot = python_version.replace(".", "")

    # create cross-platform Python before-build script to verify versions pre-build
    before_build_script = project_dir / "check_versions.py"
    before_build_script.write_text(CHECK_VERSIONS_SCRIPT)

    if utils.get_platform() == "pyodide":
        constraint_filename = f"constraints-pyodide{version_no_dot}.txt"
    else:
        constraint_filename = f"constraints-python{version_no_dot}.txt"
    constraint_file = resources.PATH / constraint_filename
    constraint_versions = get_versions_from_constraint_file(constraint_file)

    # build and test the wheels (dependency version check occurs before-build)
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_BUILD": f"[cp]p{version_no_dot}-*",
            "CIBW_BEFORE_BUILD": f"python {before_build_script.name}",
            "EXPECTED_VERSIONS": json.dumps(constraint_versions),
            **build_frontend_env_nouv,
        },
    )

    # also check that we got the right wheels
    expected_wheels = [
        w
        for w in utils.expected_wheels("spam", "0.1.0")
        if f"-cp{version_no_dot}" in w or f"-pp{version_no_dot}" in w
    ]

    assert set(actual_wheels) == set(expected_wheels)


@pytest.mark.parametrize("method", ["inline", "file"])
def test_dependency_constraints(
    method: str, tmp_path: Path, build_frontend_env_nouv: dict[str, str]
) -> None:
    if utils.get_platform() == "linux":
        pytest.skip("linux doesn't pin individual tool versions, it pins manylinux images instead")

    project_dir = tmp_path / "project"
    test_projects.new_c_project().generate(project_dir)

    tool_versions = {
        "pip": "23.1.2",
        "build": "1.2.2",
        "delocate": "0.10.3",
    }

    if method == "file":
        constraints_file = tmp_path / "constraints file.txt"
        constraints_file.write_text(
            textwrap.dedent(
                """
                pip=={pip}
                build=={build}
                delocate=={delocate}
                """.format(**tool_versions)
            )
        )
        dependency_version_option = str(constraints_file)
    elif method == "inline":
        dependency_version_option = "packages: " + " ".join(
            f"{k}=={v}" for k, v in tool_versions.items()
        )
    else:
        msg = f"Unknown method: {method}"
        raise ValueError(msg)

    skip = ""

    if (
        utils.get_platform() == "windows"
        and method == "file"
        and build_frontend_env_nouv["CIBW_BUILD_FRONTEND"] == "build"
    ):
        # GraalPy 24 fails to discover its standard library when a venv is created
        # from a virtualenv seeded executable. See
        # https://github.com/oracle/graalpython/issues/491 and remove this once
        # GraalPy 24 is dropped
        skip = "gp311*"

    # cross-platform Python script for dependency constraint checks
    before_build_script = project_dir / "check_versions.py"
    before_build_script.write_text(CHECK_VERSIONS_SCRIPT)

    # build and test the wheels (dependency version check occurs pre-build)
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_SKIP": skip,
            "CIBW_DEPENDENCY_VERSIONS": dependency_version_option,
            "CIBW_BEFORE_BUILD": f"python {before_build_script.name}",
            "EXPECTED_VERSIONS": json.dumps(tool_versions),
            **build_frontend_env_nouv,
        },
        single_python=True,
    )

    # also check that we got the right wheels
    expected_wheels = utils.expected_wheels("spam", "0.1.0", single_python=True)

    if skip == "gp*":
        # See reference to https://github.com/oracle/graalpython/issues/491
        # above
        expected_wheels = [w for w in expected_wheels if "graalpy311" not in w]

    assert set(actual_wheels) == set(expected_wheels)



import contextlib
import sys
from collections.abc import Generator
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
import setuptools._distutils.util

from cibuildwheel.ci import CIProvider, detect_ci_provider
from cibuildwheel.errors import FatalError
from cibuildwheel.platforms.windows import PythonConfiguration, setup_setuptools_cross_compile

# monkeypatching os.name is too flaky. E.g. It works on my machine, but fails in pipeline
if not sys.platform.startswith("win") and not TYPE_CHECKING:
    pytest.skip("Windows-only tests", allow_module_level=True)


@contextlib.contextmanager
def patched_environment(
    monkeypatch: pytest.MonkeyPatch, environment: dict[str, str]
) -> Generator[None, None, None]:
    with monkeypatch.context() as mp:
        for envvar, val in environment.items():
            mp.setenv(name=envvar, value=val)
        yield


def test_x86(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    environment: dict[str, str] = {}
    configuration = PythonConfiguration(version="irrelevant", identifier="cp314-win32", url=None)

    setup_setuptools_cross_compile(tmp_path, configuration, tmp_path, environment)
    with patched_environment(monkeypatch, environment):
        target_platform = setuptools._distutils.util.get_platform()

    assert environment["VSCMD_ARG_TGT_ARCH"] == "x86"
    assert target_platform == "win32"


def test_x64(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    environment: dict[str, str] = {}
    configuration = PythonConfiguration(
        version="irrelevant", identifier="cp313-win_amd64", url=None
    )

    setup_setuptools_cross_compile(tmp_path, configuration, tmp_path, environment)
    with patched_environment(monkeypatch, environment):
        target_platform = setuptools._distutils.util.get_platform()

    assert environment["VSCMD_ARG_TGT_ARCH"] == "x64"
    assert target_platform == "win-amd64"


@pytest.mark.skipif(
    detect_ci_provider() == CIProvider.azure_pipelines, reason="arm64 not recognised on azure"
)
def test_arm(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    environment: dict[str, str] = {}
    configuration = PythonConfiguration(
        version="irrelevant", identifier="cp312-win_arm64", url=None
    )

    setup_setuptools_cross_compile(tmp_path, configuration, tmp_path, environment)
    with patched_environment(monkeypatch, environment):
        target_platform = setuptools._distutils.util.get_platform()

    assert environment["VSCMD_ARG_TGT_ARCH"] == "arm64"
    assert target_platform == "win-arm64"


def test_env_set(tmp_path: Path) -> None:
    environment = {"VSCMD_ARG_TGT_ARCH": "x64"}
    configuration = PythonConfiguration(version="irrelevant", identifier="cp314t-win32", url=None)

    with pytest.raises(FatalError, match="VSCMD_ARG_TGT_ARCH"):
        setup_setuptools_cross_compile(tmp_path, configuration, tmp_path, environment)


def test_env_blank(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    environment = {"VSCMD_ARG_TGT_ARCH": ""}
    configuration = PythonConfiguration(version="irrelevant", identifier="cp312-win32", url=None)

    setup_setuptools_cross_compile(tmp_path, configuration, tmp_path, environment)
    with patched_environment(monkeypatch, environment):
        target_platform = setuptools._distutils.util.get_platform()

    assert environment["VSCMD_ARG_TGT_ARCH"] == "x86"
    assert target_platform == "win32"



import textwrap
from pathlib import Path

import pytest

from . import test_projects, utils

project_with_ssl_tests = test_projects.new_c_project(
    setup_py_add=textwrap.dedent(
        r"""
            import ssl, time, urllib.request, urllib.error

            def check_https(context=None):
                # google hosts this endpoint that returns a 204 No Content, it's used for
                # connectivity checks in Android & Chrome
                url = "https://google.com/generate_204"

                for i in range(5):
                    try:
                        urllib.request.urlopen(url, context=context, timeout=5)
                        return
                    except (OSError, urllib.error.URLError) as e:
                        print(f"Attempt {i+1}: Could not connect to {url}: {e}")
                        time.sleep(2 ** i)  # Backoff: 1s, 2s, 4s, 8s...

                raise ConnectionError(f"Could not connect to {url} after retries.")

            check_https()
            check_https(context=ssl.SSLContext(ssl.PROTOCOL_TLSv1_2))
        """
    )
)


@pytest.mark.flaky(reruns=2)
def test(tmp_path: Path) -> None:
    # this test checks that SSL is working in the build environment using
    # some checks in setup.py.
    project_dir = tmp_path / "project"
    project_with_ssl_tests.generate(project_dir)

    actual_wheels = utils.cibuildwheel_run(project_dir)

    expected_wheels = utils.expected_wheels("spam", "0.1.0")
    assert set(actual_wheels) == set(expected_wheels)



#!/usr/bin/env python3

import sys
import textwrap
from pathlib import Path

import click

from cibuildwheel.extra import get_pyodide_xbuildenv_info


@click.command()
@click.argument(
    "pyodide-version",
    type=str,
)
@click.option(
    "--output-file",
    type=click.Path(),
    default=None,
    help="Output file to write the constraints to. If not provided, the constraints will be printed to stdout.",
)
def generate_pyodide_constraints(pyodide_version: str, output_file: str | None = None) -> None:
    """
    Generate constraints for a specific Pyodide version. The constraints are
    generated based on the Pyodide version's xbuildenv info, which is retrieved
    from the Pyodide repository.

    These constraints should then be 'pinned' using `uv pip compile`.

    Example usage:

        bin/generate_pyodide_constraints.py 0.27.0
    """
    xbuildenv_info = get_pyodide_xbuildenv_info()
    try:
        pyodide_version_xbuildenv_info = xbuildenv_info["releases"][pyodide_version]
    except KeyError as e:
        msg = f"Pyodide version {pyodide_version} not found in xbuildenv info. Versions available: {', '.join(xbuildenv_info['releases'].keys())}"
        raise click.BadParameter(msg) from e

    pyodide_build_min_version = pyodide_version_xbuildenv_info.get("min_pyodide_build_version")
    pyodide_build_max_version = pyodide_version_xbuildenv_info.get("max_pyodide_build_version")

    pyodide_build_specifier_parts: list[str] = []

    if pyodide_build_min_version:
        pyodide_build_specifier_parts.append(f">={pyodide_build_min_version}")
    if pyodide_build_max_version:
        pyodide_build_specifier_parts.append(f"<={pyodide_build_max_version}")

    pyodide_build_specifier = ",".join(pyodide_build_specifier_parts)

    constraints_txt = textwrap.dedent(f"""
        pip
        build[virtualenv]
        pyodide-build{pyodide_build_specifier}
        click<8.2
    """)

    if output_file is None:
        print(constraints_txt)
    else:
        Path(output_file).write_text(constraints_txt)
        print(f"Constraints written to {output_file}", file=sys.stderr)


if __name__ == "__main__":
    generate_pyodide_constraints()



import dataclasses
import shlex
import typing
from collections.abc import Sequence
from typing import Literal, Self, get_args

from cibuildwheel.logger import log
from cibuildwheel.util.helpers import parse_key_value_string

BuildFrontendName = Literal["pip", "build", "build[uv]", "uv"]


@dataclasses.dataclass(frozen=True)
class BuildFrontendConfig:
    name: BuildFrontendName
    args: Sequence[str] = ()

    @classmethod
    def from_config_string(cls, config_string: str) -> Self:
        config_dict = parse_key_value_string(config_string, ["name"], ["args"])
        name = " ".join(config_dict["name"])
        if name not in get_args(BuildFrontendName):
            names = ", ".join(repr(n) for n in get_args(BuildFrontendName))
            msg = f"Unrecognised build frontend {name!r}, must be one of {names}"
            raise ValueError(msg)

        name = typing.cast("BuildFrontendName", name)

        args = config_dict.get("args") or []
        return cls(name=name, args=args)

    def options_summary(self) -> str | dict[str, str]:
        if not self.args:
            return self.name
        else:
            return {"name": self.name, "args": repr(self.args)}


def _get_verbosity_flags(level: int, frontend: BuildFrontendName, *, py38: bool) -> list[str]:
    if level < 0:
        if frontend.startswith("build") and py38:
            msg = f"build_verbosity {level} is not supported for {frontend} frontend. Ignoring."
            log.warning(msg)
            return []
        return ["-" + -level * "q"]

    if level > 0:
        if frontend == "pip":
            return ["-" + level * "v"]
        if level > 1:
            return ["-" + (level - 1) * "v"]

    return []


def _split_config_settings(config_settings: str) -> list[str]:
    config_settings_list = shlex.split(config_settings)
    return [f"-C{setting}" for setting in config_settings_list]


# Based on build.__main__.main.
def parse_config_settings(config_settings_str: str) -> dict[str, str | list[str]]:
    config_settings: dict[str, str | list[str]] = {}
    for arg in shlex.split(config_settings_str):
        setting, _, value = arg.partition("=")
        existing_value = config_settings.get(setting)
        if existing_value is None:
            config_settings[setting] = value
        elif isinstance(existing_value, str):
            config_settings[setting] = [existing_value, value]
        else:
            existing_value.append(value)

    return config_settings


def get_build_frontend_extra_flags(
    build_frontend: BuildFrontendConfig, verbosity_level: int, config_settings: str, *, py38: bool
) -> list[str]:
    return [
        *_split_config_settings(config_settings),
        *build_frontend.args,
        *_get_verbosity_flags(verbosity_level, build_frontend.name, py38=py38),
    ]



import platform as platform_module
import shutil
import sys

import pytest

import cibuildwheel.architecture
from cibuildwheel.architecture import Architecture, arch_synonym
from cibuildwheel.typing import PlatformName


@pytest.fixture(
    params=[
        pytest.param(("linux", "linux", "x86_64", "64"), id="linux-64"),
        pytest.param(("linux", "linux", "i686", "32"), id="linux-32"),
        pytest.param(("linux", "linux", "aarch64", "arm"), id="linux-arm"),
        pytest.param(("macos", "darwin", "x86_64", "64"), id="macos-64"),
        pytest.param(("macos", "darwin", "arm64", "arm"), id="macos-arm"),
        pytest.param(("windows", "win32", "x86", "32"), id="windows-32"),
        pytest.param(("windows", "win32", "AMD64", "64"), id="windows-64"),
        pytest.param(("windows", "win32", "ARM64", "arm"), id="windows-arm"),
    ]
)
def platform_machine(
    request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch
) -> tuple[str, str]:
    platform_name, platform_value, machine_value, machine_name = request.param
    monkeypatch.setattr(sys, "platform", platform_value)
    monkeypatch.setattr(platform_module, "machine", lambda: machine_value)
    monkeypatch.setattr(cibuildwheel.architecture, "_check_aarch32_el0", lambda: True)
    return platform_name, machine_name


def test_arch_auto(platform_machine: tuple[str, str]) -> None:
    _, machine_name = platform_machine

    arch_set = Architecture.auto_archs("linux")
    expected = {
        "32": {Architecture.i686},
        "64": {Architecture.x86_64},
        "arm": {Architecture.aarch64},
    }
    assert arch_set == expected[machine_name]

    arch_set = Architecture.auto_archs("macos")
    expected = {"32": set(), "64": {Architecture.x86_64}, "arm": {Architecture.arm64}}
    assert arch_set == expected[machine_name]

    arch_set = Architecture.auto_archs("windows")
    expected = {
        "32": {Architecture.x86},
        "64": {Architecture.AMD64, Architecture.x86},
        "arm": {Architecture.ARM64},
    }
    assert arch_set == expected[machine_name]


def test_arch_auto64(platform_machine: tuple[str, str]) -> None:
    _, machine_name = platform_machine

    arch_set = Architecture.parse_config("auto64", "linux")
    expected = {"32": set(), "64": {Architecture.x86_64}, "arm": {Architecture.aarch64}}
    assert arch_set == expected[machine_name]

    arch_set = Architecture.parse_config("auto64", "macos")
    expected = {"32": set(), "64": {Architecture.x86_64}, "arm": {Architecture.arm64}}
    assert arch_set == expected[machine_name]

    arch_set = Architecture.parse_config("auto64", "windows")
    expected = {"32": set(), "64": {Architecture.AMD64}, "arm": {Architecture.ARM64}}
    assert arch_set == expected[machine_name]


def test_arch_auto32(platform_machine: tuple[str, str]) -> None:
    _, machine_name = platform_machine

    arch_set = Architecture.parse_config("auto32", "linux")
    expected = {"32": {Architecture.i686}, "64": {Architecture.i686}, "arm": {Architecture.armv7l}}
    assert arch_set == expected[machine_name]

    arch_set = Architecture.parse_config("auto32", "macos")
    assert arch_set == set()

    arch_set = Architecture.parse_config("auto32", "windows")
    expected = {"32": {Architecture.x86}, "64": {Architecture.x86}, "arm": set()}
    assert arch_set == expected[machine_name]


def test_arch_auto_no_aarch32(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(sys, "platform", "linux")
    monkeypatch.setattr(platform_module, "machine", lambda: "aarch64")
    monkeypatch.setattr(shutil, "which", lambda *args, **kwargs: None)

    arch_set = Architecture.parse_config("auto", "linux")
    assert arch_set == {Architecture.aarch64}

    arch_set = Architecture.parse_config("auto64", "linux")
    assert arch_set == {Architecture.aarch64}

    monkeypatch.setattr(cibuildwheel.architecture, "_check_aarch32_el0", lambda: True)
    arch_set = Architecture.parse_config("auto32", "linux")
    assert arch_set == {Architecture.armv7l}

    monkeypatch.setattr(cibuildwheel.architecture, "_check_aarch32_el0", lambda: False)
    arch_set = Architecture.parse_config("auto32", "linux")
    assert arch_set == set()


def test_arch_native_on_ios(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(sys, "platform", "darwin")
    monkeypatch.setattr(platform_module, "machine", lambda: "arm64")
    arch_set = Architecture.parse_config("native", platform="ios")
    assert arch_set == {Architecture.arm64_iphonesimulator}


def test_arch_auto_on_ios(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(sys, "platform", "darwin")
    monkeypatch.setattr(platform_module, "machine", lambda: "arm64")
    arch_set = Architecture.parse_config("auto", platform="ios")
    assert arch_set == {Architecture.arm64_iphonesimulator, Architecture.arm64_iphoneos}


@pytest.mark.parametrize(
    ("arch", "from_platform", "to_platform", "expected"),
    [
        ("x86_64", "linux", "macos", "x86_64"),
        ("x86_64", "macos", "linux", "x86_64"),
        ("x86_64", "linux", "windows", "AMD64"),
        ("AMD64", "windows", "linux", "x86_64"),
        ("x86_64", "linux", "nonexistent", "x86_64"),
        ("x86_64", "nonexistent", "linux", "x86_64"),
        ("nonexistent", "linux", "windows", "nonexistent"),
        ("x86", "windows", "macos", None),
    ],
)
def test_arch_synonym(
    arch: str, from_platform: PlatformName, to_platform: PlatformName, expected: str | None
) -> None:
    assert arch_synonym(arch, from_platform, to_platform) == expected



#!/usr/bin/env python3

# /// script
# dependencies = [
#   "click",
#   "pyyaml",
#   "pygithub",
# ]
# ///

"""
Convert a yaml project list into a nice table.

Suggested usage:

    ./bin/projects.py docs/data/projects.yml --online --auth $GITHUB_API_TOKEN
    git diff
"""

import builtins
import functools
import textwrap
import urllib.request
import xml.dom.minidom
from collections.abc import Iterable, Mapping, Sequence
from datetime import UTC, datetime
from io import StringIO
from pathlib import Path
from typing import Any, Self, TextIO

import click
import yaml
from github import Github, GithubException

ICONS = (
    "github",
    "azurepipelines",
    "circleci",
    "gitlab",
    "travisci",
    "cirrusci",
    "windows",
    "apple",
    "linux",
)


class Project:
    NAME: int = 0

    def __init__(self, config: Mapping[str, Any], github: Github | None = None):
        try:
            self.name: str = config["name"]
            self.gh: str = config["gh"]
        except KeyError:
            print("Invalid config, needs at least gh and name!", config)
            raise

        self.stars_repo: str = config.get("stars", self.gh)
        self.notes: str = config.get("notes", "")
        self.ci: list[str] = config.get("ci", [])
        self.os: list[str] = config.get("os", [])

        self.online = github is not None
        if github is not None:
            try:
                repo = github.get_repo(self.stars_repo)
            except GithubException:
                print(f"Broken: {self.stars_repo}")
                raise

            self.num_stars: int = repo.stargazers_count
            self.pushed_at = repo.pushed_at
            if not self.notes:
                notes = repo.description
                if repo.description:
                    self.notes = notes
        else:
            self.num_stars = 0
            self.pushed_at = datetime.now(UTC)

        name_len = len(self.name) + 4
        self.__class__.NAME = max(self.__class__.NAME, name_len)

    def __lt__(self, other: Self) -> bool:
        if self.online:
            return self.num_stars < other.num_stars
        else:
            return self.name < other.name

    @classmethod
    def header(cls) -> str:
        return textwrap.dedent(
            f"""\
                | {"Name":{cls.NAME}} | CI | OS | Notes |
                |{"":-^{cls.NAME + 2}}|----|----|:------|"""
        )

    @property
    def namelink(self) -> str:
        return f"[{self.name}][]"

    @property
    def starslink(self) -> str:
        return f"![{self.name} stars][]"

    @property
    def url(self) -> str:
        return f"https://github.com/{self.gh}"

    @property
    def ci_icons(self) -> str:
        return " ".join(f"![{icon} icon][]" for icon in self.ci)

    @property
    def os_icons(self) -> str:
        return " ".join(f"![{icon} icon][]" for icon in self.os)

    def table_row(self) -> str:
        notes = self.notes.replace("\n", " ")
        return f"| {self.namelink: <{self.NAME}} | {self.ci_icons} | {self.os_icons} | {notes} |"

    def links(self) -> str:
        return f"[{self.name}]: {self.url}"

    def info(self) -> str:
        days = (datetime.now(UTC) - self.pushed_at).days
        return f"<!-- {self.name}: {self.num_stars}, last pushed {days} days ago -->"


def fetch_icon(icon_name: str) -> None:
    url = f"https://cdn.jsdelivr.net/npm/simple-icons@v4/icons/{icon_name}.svg"
    with urllib.request.urlopen(url) as f:
        original_svg_data = f.read()

    document = xml.dom.minidom.parseString(original_svg_data)
    svgElement = document.documentElement
    assert svgElement is not None
    assert svgElement.nodeName == "svg"
    svgElement.setAttribute("width", "16px")
    svgElement.setAttribute("fill", "#606060")

    icon_path = path_for_icon(icon_name)
    icon_path.parent.mkdir(parents=True, exist_ok=True)

    with open(path_for_icon(icon_name), "w") as f:
        f.write(svgElement.toxml())


def path_for_icon(icon_name: str, relative_to: Path | None = None) -> Path:
    from_path = Path.cwd() if relative_to is None else relative_to.parent

    absolute_path = PROJECT_ROOT / "docs" / "data" / "readme_icons" / f"{icon_name}.svg"

    return absolute_path.resolve().relative_to(from_path.resolve())


def get_projects(
    config: Iterable[Mapping[str, Any]],
    *,
    online: bool = True,
    auth: str | None = None,
) -> list[Project]:
    if online:
        for icon in ICONS:
            fetch_icon(icon)

    github = Github(auth) if online else None

    return sorted((Project(item, github) for item in config), reverse=online)


def render_projects(
    projects: Sequence[Project], *, dest_path: Path, include_info: bool = True
) -> str:
    io = StringIO()
    print = functools.partial(builtins.print, file=io)

    print(Project.header())

    for project in projects:
        print(project.table_row())

    print()
    for project in projects:
        print(project.links())

    print()
    for icon in ICONS:
        print(f"[{icon} icon]: {path_for_icon(icon, relative_to=dest_path).as_posix()}")

    if include_info:
        print()
        for project in projects:
            print(project.info())

    return io.getvalue()


def insert_projects_table(
    file: Path,
    *,
    projects: Sequence[Project],
    input_filename: str,
    include_info: bool = True,
) -> None:
    text = file.read_text()
    projects_table = render_projects(projects, include_info=include_info, dest_path=file)

    start_str = "<!-- START bin/projects.py -->\n"

    start = text.find(start_str)
    assert start != -1

    end = text.find("<!-- END bin/projects.py -->\n")
    assert end != -1

    generated_note = f"<!-- this section is generated by bin/projects.py. Don't edit it directly, instead, edit {input_filename} -->"
    new_text = (
        f"{text[: start + len(start_str)]}\n{generated_note}\n\n{projects_table}\n{text[end:]}"
    )

    file.write_text(new_text)


PROJECT_ROOT = Path(__file__).parent / ".."
README_FILE = PROJECT_ROOT / "README.md"
DOCS_PAGE = PROJECT_ROOT / "docs" / "working-examples.md"


@click.command(help="Try ./bin/projects.py docs/data/projects.yml")
@click.argument("input", type=click.File("r"))
@click.option("--online/--no-online", default=True, help="Get info from GitHub")
@click.option("--auth", help="GitHub authentication token")
@click.option(
    "--dry-run", default=False, help="Print the output, rather than writing it to files in the repo"
)
def projects(
    input: TextIO,
    online: bool,
    auth: str | None,
    dry_run: bool,
) -> None:
    config = yaml.safe_load(input)
    projects = get_projects(config, online=online, auth=auth)

    if dry_run:
        output = render_projects(projects, dest_path=README_FILE)
        print(output)
    else:
        insert_projects_table(
            README_FILE, projects=projects[:10], input_filename=input.name, include_info=False
        )
        insert_projects_table(
            DOCS_PAGE, projects=projects, input_filename=input.name, include_info=False
        )


if __name__ == "__main__":
    projects()



#!/usr/bin/env python3


import os
import subprocess
import sys
import textwrap
import time
from pathlib import Path

import click


def shell(cmd: str, *, check: bool, **kwargs: object) -> subprocess.CompletedProcess[str]:
    return subprocess.run([cmd], shell=True, check=check, **kwargs)  # type: ignore[call-overload, no-any-return]


def git_repo_has_changes() -> bool:
    unstaged_changes: bool = shell("git diff-index --quiet HEAD --", check=False).returncode != 0
    staged_changes: bool = (
        shell("git diff-index --quiet --cached HEAD --", check=False).returncode != 0
    )
    return unstaged_changes or staged_changes


@click.command()
def main() -> None:
    project_root = Path(__file__).parent / ".."
    os.chdir(project_root)

    if git_repo_has_changes():
        print("Your git repo has uncommitted changes. Commit or stash before continuing.")
        sys.exit(1)

    previous_branch = shell(
        "git rev-parse --abbrev-ref HEAD", check=True, capture_output=True, encoding="utf8"
    ).stdout.strip()

    shell("git fetch origin", check=True)

    timestamp = time.strftime("%Y-%m-%dT%H-%M-%S", time.gmtime())
    branch_name = f"update-constraints-{timestamp}"

    shell(f"git checkout -b {branch_name} origin/main", check=True)

    try:
        shell("bin/update_dependencies.py", check=True)

        if not git_repo_has_changes():
            print("Done: no constraint updates required.")
            return

        shell('git commit -a -m "Update dependencies"', check=True)
        body = textwrap.dedent(
            f"""
            Update the versions of our dependencies.

            PR generated by `{Path(__file__).name}`.
            """
        )
        subprocess.run(
            [
                "gh",
                "pr",
                "create",
                "--repo=pypa/cibuildwheel",
                "--base=main",
                "--title=Update dependencies",
                f"--body='{body}'",
            ],
            check=True,
        )

        print("Done.")
    finally:
        # remove any local changes
        shell("git checkout -- .", check=True)
        shell(f"git checkout {previous_branch}", check=True)
        shell(f"git branch -D --force {branch_name}", check=True)


if __name__ == "__main__":
    main.main(standalone_mode=True)



import shlex
from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from pathlib import Path, PurePath
from typing import Literal, Self, TypeVar

from packaging.utils import parse_wheel_filename

from cibuildwheel.util import resources
from cibuildwheel.util.cmd import call
from cibuildwheel.util.helpers import parse_key_value_string, unwrap


@dataclass(kw_only=True)
class DependencyConstraints:
    base_file_path: Path | None = None
    packages: list[str] = field(default_factory=list)

    def __post_init__(self) -> None:
        if self.packages and self.base_file_path is not None:
            msg = "Cannot specify both a file and packages in the dependency constraints"
            raise ValueError(msg)

        if self.base_file_path is not None:
            if not self.base_file_path.exists():
                msg = f"Dependency constraints file not found: {self.base_file_path}"
                raise FileNotFoundError(msg)
            self.base_file_path = self.base_file_path.resolve()

    @classmethod
    def pinned(cls) -> Self:
        return cls(base_file_path=resources.CONSTRAINTS)

    @classmethod
    def latest(cls) -> Self:
        return cls()

    @classmethod
    def from_config_string(cls, config_string: str) -> Self:
        if config_string == "pinned":
            return cls.pinned()

        if config_string == "latest" or not config_string:
            return cls.latest()

        if config_string.startswith(("file:", "packages:")):
            # we only do the table-style parsing if it looks like a table,
            # because this option used to be only a file path. We don't want
            # to break existing configurations, whose file paths might include
            # special characters like ':' or ' ', which would require quoting
            # if they were to be passed as a parse_key_value_string positional
            # argument.
            return cls.from_table_style_config_string(config_string)

        return cls(base_file_path=Path(config_string))

    @classmethod
    def from_table_style_config_string(cls, config_string: str) -> Self:
        config_dict = parse_key_value_string(config_string, kw_arg_names=["file", "packages"])
        files = config_dict.get("file")
        packages = config_dict.get("packages") or []

        if files and packages:
            msg = "Cannot specify both a file and packages in dependency-versions"
            raise ValueError(msg)

        if files:
            if len(files) > 1:
                msg = unwrap("""
                    Only one file can be specified in dependency-versions.
                    If you intended to pass only one, perhaps you need to quote the path?
                """)
                raise ValueError(msg)

            return cls(base_file_path=Path(files[0]))

        return cls(packages=packages)

    def get_for_python_version(
        self, *, version: str, variant: Literal["python", "pyodide"] = "python", tmp_dir: Path
    ) -> Path | None:
        if self.packages:
            constraint_file = tmp_dir / "constraints.txt"
            constraint_file.write_text("\n".join(self.packages))
            return constraint_file

        if self.base_file_path is not None:
            version_parts = version.split(".")

            # try to find a version-specific dependency file e.g. if
            # ./constraints.txt is the base, look for ./constraints-python36.txt
            specific_stem = (
                self.base_file_path.stem + f"-{variant}{version_parts[0]}{version_parts[1]}"
            )
            specific_name = specific_stem + self.base_file_path.suffix
            specific_file_path = self.base_file_path.with_name(specific_name)

            if specific_file_path.exists():
                return specific_file_path
            else:
                return self.base_file_path

        return None

    def options_summary(self) -> str | dict[str, str]:
        if self == DependencyConstraints.pinned():
            return "pinned"
        elif self.packages:
            return {"packages": " ".join(shlex.quote(p) for p in self.packages)}
        elif self.base_file_path is not None:
            return self.base_file_path.name
        else:
            return "latest"


def get_pip_version(env: Mapping[str, str]) -> str:
    versions_output_text = call(
        "python", "-m", "pip", "freeze", "--all", capture_stdout=True, env=env
    )
    (pip_version,) = (
        version[5:]
        for version in versions_output_text.strip().splitlines()
        if version.startswith("pip==")
    )
    return pip_version


T = TypeVar("T", bound=PurePath)


def find_compatible_wheel(wheels: Sequence[T], identifier: str) -> T | None:
    """
    Finds a wheel with an abi3 or a none ABI tag in `wheels` compatible with the Python interpreter
    specified by `identifier` that is previously built.
    """

    interpreter, platform = identifier.split("-", 1)
    interpreter = interpreter.split("_")[0]
    free_threaded = interpreter.endswith("t")
    if free_threaded:
        interpreter = interpreter[:-1]
    for wheel in wheels:
        _, _, _, tags = parse_wheel_filename(wheel.name)
        for tag in tags:
            if tag.abi == "abi3" and not free_threaded:
                # ABI3 wheels must start with cp3 for impl and tag
                if not (interpreter.startswith("cp3") and tag.interpreter.startswith("cp3")):
                    continue
            elif tag.abi == "none":
                # CPythonless wheels must include py3 tag
                if tag.interpreter[:3] != "py3":
                    continue
            else:
                # Other types of wheels are not detected, this is looking for previously built wheels.
                continue

            if tag.interpreter != "py3" and int(tag.interpreter[3:]) > int(interpreter[3:]):
                # If a minor version number is given, it has to be lower than the current one.
                continue

            if platform.startswith(("manylinux", "musllinux", "macosx", "android", "ios")):
                # On these platforms the wheel tag includes a platform version number, which we
                # should ignore.
                os_, arch = platform.split("_", 1)
                if not tag.platform.startswith(os_):
                    continue
                if not tag.platform.endswith(f"_{arch}"):
                    continue
            elif platform.startswith("pyodide"):
                # each Pyodide version has its own platform tag
                continue
            # Windows should exactly match
            elif tag.platform != platform:
                continue

            # If all the filters above pass, then the wheel is a previously built compatible wheel.
            return wheel

    return None



import re
import tomllib
from pathlib import Path

import pytest
import validate_pyproject.api

from cibuildwheel.util import resources

DIR = Path(__file__).parent.resolve()


@pytest.fixture(scope="session")
def validator() -> validate_pyproject.api.Validator:
    """
    Reuse the validator for all tests, to keep unit tests fast.
    """
    return validate_pyproject.api.Validator()


def test_validate_default_schema(validator: validate_pyproject.api.Validator) -> None:
    with resources.DEFAULTS.open("rb") as f:
        example = tomllib.load(f)

    assert validator(example) is not None


def test_validate_container_engine(validator: validate_pyproject.api.Validator) -> None:
    """
    This test checks container engine can be overridden - it used to be a
    global option but is now a build option.
    """

    example = tomllib.loads(
        """
        [tool.cibuildwheel]
        container-engine = "docker"

        [tool.cibuildwheel.linux]
        container-engine = "docker"

        [[tool.cibuildwheel.overrides]]
        select = "*_x86_64"
        container-engine = "docker; create_args: --platform linux/arm64/v8"
        """
    )

    assert validator(example) is not None


@pytest.mark.parametrize("platform", ["macos", "windows"])
def test_validate_bad_container_engine(
    validator: validate_pyproject.api.Validator, platform: str
) -> None:
    """
    container-engine is not a valid option for macos or windows
    """
    example = tomllib.loads(
        f"""
        [tool.cibuildwheel.{platform}]
        container-engine = "docker"
        """
    )

    with pytest.raises(validate_pyproject.error_reporting.ValidationError):
        validator(example)


def test_overrides_select(validator: validate_pyproject.api.Validator) -> None:
    example = tomllib.loads(
        """
        [[tool.cibuildwheel.overrides]]
        select = "somestring"
        repair-wheel-command = "something"
        """
    )

    assert validator(example) is not None


def test_overrides_no_select(validator: validate_pyproject.api.Validator) -> None:
    example = tomllib.loads(
        """
        [[tool.cibuildwheel.overrides]]
        repair-wheel-command = "something"
        """
    )

    with pytest.raises(validate_pyproject.error_reporting.ValidationError):
        validator(example)


def test_overrides_only_select(validator: validate_pyproject.api.Validator) -> None:
    example = tomllib.loads(
        """
        [[tool.cibuildwheel.overrides]]
        select = "somestring"
        """
    )

    with pytest.raises(validate_pyproject.error_reporting.ValidationError):
        validator(example)


def test_overrides_valid_inherit(validator: validate_pyproject.api.Validator) -> None:
    example = tomllib.loads(
        """
        [[tool.cibuildwheel.overrides]]
        inherit.repair-wheel-command = "append"
        select = "somestring"
        repair-wheel-command = ["something"]
        """
    )

    assert validator(example) is not None


def test_overrides_invalid_inherit(validator: validate_pyproject.api.Validator) -> None:
    example = tomllib.loads(
        """
        [[tool.cibuildwheel.overrides]]
        inherit.something = "append"
        select = "somestring"
        repair-wheel-command = "something"
        """
    )

    with pytest.raises(validate_pyproject.error_reporting.ValidationError):
        validator(example)


def test_overrides_invalid_inherit_value(validator: validate_pyproject.api.Validator) -> None:
    example = tomllib.loads(
        """
        [[tool.cibuildwheel.overrides]]
        inherit.repair-wheel-command = "nothing"
        select = "somestring"
        repair-wheel-command = "something"
        """
    )

    with pytest.raises(validate_pyproject.error_reporting.ValidationError):
        validator(example)


def test_docs_examples(validator: validate_pyproject.api.Validator) -> None:
    """
    Parse out all the configuration examples, build valid TOML out of them, and
    make sure they pass.
    """

    expr = re.compile(
        r"""
!!! tab examples "pyproject.toml"
\s*
\s*```toml
(.*?)```""",
        re.MULTILINE | re.DOTALL,
    )

    txt = DIR.parent.joinpath("docs/options.md").read_text()

    blocks: list[str] = []
    for match in expr.finditer(txt):
        lines = (line.strip() for line in match.group(1).strip().splitlines() if line.strip())
        block: list[str] = []
        header = ""
        for line in lines:
            if line.startswith(("[tool.cibuildwheel", "[[tool.cibuildwheel")):
                header = line
            elif line.startswith("#"):
                if block:
                    blocks.append("\n".join([header, *block]))
                    block = []
            elif " = " in line and any(x.startswith(line.partition(" = ")[0]) for x in block):
                blocks.append("\n".join([header, *block]))
                block = [line]
            else:
                block.append(line)
        blocks.append("\n".join([header, *block]))

    for example_txt in blocks:
        print(example_txt)
        print()
        example = tomllib.loads(example_txt)

        assert validator(example) is not None



from pathlib import Path

import pytest

from cibuildwheel.logger import BuildInfo, Logger
from cibuildwheel.options import CommandLineArguments, Options

OPTIONS_DEFAULTS = Options("linux", CommandLineArguments.defaults(), {}, defaults=True)
FILE = Path(__file__)


def test_printout_wheels(capsys: pytest.CaptureFixture[str]) -> None:
    log = Logger()
    log.fold_mode = "disabled"
    log.colors_enabled = False

    with log.print_summary(options=OPTIONS_DEFAULTS):
        # the number of BuildInfo with & without filename shall be different for this test
        log.summary = [
            BuildInfo(identifier="id1", filename=None, duration=3),
            BuildInfo(identifier="id2", filename=FILE, duration=2),
            BuildInfo(identifier="id3", filename=FILE, duration=3),
        ]

    captured = capsys.readouterr()
    assert captured.err == ""

    assert "id1" in captured.out
    assert "id2" in captured.out
    assert "id3" in captured.out
    assert "2 wheels produced in" in captured.out
    assert "SHA256=" in captured.out


def test_no_printout_on_error(capsys: pytest.CaptureFixture[str]) -> None:
    log = Logger()
    with pytest.raises(RuntimeError), log.print_summary(options=OPTIONS_DEFAULTS):
        raise RuntimeError()

    captured = capsys.readouterr()
    assert captured.err == ""
    assert captured.out == ""



import os
import re
from enum import Enum

from cibuildwheel.util.helpers import strtobool

ANSI_CODE_REGEX = re.compile(r"(\033\[[0-9;]*m)")


class CIProvider(Enum):
    # official support
    travis_ci = "travis"
    circle_ci = "circle_ci"
    azure_pipelines = "azure_pipelines"
    github_actions = "github_actions"
    gitlab = "gitlab"
    cirrus_ci = "cirrus_ci"

    # unofficial support
    appveyor = "appveyor"

    other = "other"


def detect_ci_provider() -> CIProvider | None:
    if "TRAVIS" in os.environ:
        return CIProvider.travis_ci
    elif "APPVEYOR" in os.environ:
        return CIProvider.appveyor
    elif "CIRCLECI" in os.environ:
        return CIProvider.circle_ci
    elif "AZURE_HTTP_USER_AGENT" in os.environ:
        return CIProvider.azure_pipelines
    elif "GITHUB_ACTIONS" in os.environ:
        return CIProvider.github_actions
    elif "GITLAB_CI" in os.environ:
        return CIProvider.gitlab
    elif "CIRRUS_CI" in os.environ:
        return CIProvider.cirrus_ci
    elif strtobool(os.environ.get("CI", "false")):
        return CIProvider.other
    else:
        return None


def fix_ansi_codes_for_github_actions(text: str) -> str:
    """
    Github Actions forgets the current ANSI style on every new line. This
    function repeats the current ANSI style on every new line.
    """

    ansi_codes: list[str] = []
    output = ""

    for line in text.splitlines(keepends=True):
        # add the current ANSI codes to the beginning of the line
        output += "".join(ansi_codes) + line

        # split the line at each ANSI code
        parts = ANSI_CODE_REGEX.split(line)
        # if there are any ANSI codes, save them
        if len(parts) > 1:
            # iterate over the ANSI codes in this line
            for code in parts[1::2]:
                if code == "\033[0m":
                    # reset the list of ANSI codes when the clear code is found
                    ansi_codes = []
                else:
                    ansi_codes.append(code)

    return output


def filter_ansi_codes(text: str, /) -> str:
    """
    Remove ANSI codes from text.
    """

    return ANSI_CODE_REGEX.sub("", text)
